from typing import Any

import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import GeoDataModule, NonGeoDataModule
from torchgeo.samplers import GridGeoSampler, RandomBatchGeoSampler
from torchgeo.transforms import AugmentationSequential

from terratorch.datasets import FireScarsHLS, FireScarsNonGeo, FireScarsSegmentationMask

MEANS = [
    0.033349706741586264,
    0.05701185520536176,
    0.05889748132001316,
    0.2323245113436119,
    0.1972854853760658,
    0.11944914225186566,
]

STDS = [
    0.02269135568823774,
    0.026807560223070237,
    0.04004109844362779,
    0.07791732423672691,
    0.08708738838140137,
    0.07241979477437814,
]


class FireScarsNonGeoDataModule(NonGeoDataModule):
    """NonGeo Fire Scars data module implementation"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(FireScarsNonGeo, 16, 8, **kwargs)
        # applied for training
        self.train_aug = AugmentationSequential(
            K.Normalize(MEANS, STDS),
            K.RandomCrop((224, 224)),
            data_keys=["image", "mask"],
        )
        self.aug = AugmentationSequential(K.Normalize(MEANS, STDS), data_keys=["image", "mask"])

    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(
                "/dccstor/geofm-finetuning/fire-scars/finetune-data/6_bands_no_replant_extended/training/"
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                "/dccstor/geofm-finetuning/fire-scars/finetune-data/6_bands_no_replant_extended/validation/"
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                "/dccstor/geofm-finetuning/fire-scars/finetune-data/6_bands_no_replant_extended/validation/"
            )


class FireScarsDataModule(GeoDataModule):
    """Geo Fire Scars data module implementation that merges input data with ground truth segmentation masks."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(FireScarsSegmentationMask, 4, 224, 100, 0, **kwargs)
        self.train_aug = AugmentationSequential(K.RandomCrop(224, 224), K.Normalize(MEANS, STDS))
        self.aug = AugmentationSequential(K.Normalize(MEANS, STDS))

    def setup(self, stage: str) -> None:
        self.images = FireScarsHLS(
            "/dccstor/geofm-finetuning/fire-scars/finetune-data/6_bands_no_replant_extended/training/"
        )
        self.labels = FireScarsSegmentationMask(
            "/dccstor/geofm-finetuning/fire-scars/finetune-data/6_bands_no_replant_extended/training/"
        )
        self.dataset = self.images & self.labels
        self.train_aug = AugmentationSequential(K.RandomCrop(224, 224), K.normalize())

        self.images_test = FireScarsHLS(
            "/dccstor/geofm-finetuning/fire-scars/finetune-data/6_bands_no_replant_extended/validation/"
        )
        self.labels_test = FireScarsSegmentationMask(
            "/dccstor/geofm-finetuning/fire-scars/finetune-data/6_bands_no_replant_extended/validation/"
        )
        self.val_dataset = self.images_test & self.labels_test

        if stage in ["fit"]:
            self.train_batch_sampler = RandomBatchGeoSampler(self.dataset, self.patch_size, self.batch_size, None)
        if stage in ["fit", "validate"]:
            self.val_sampler = GridGeoSampler(self.val_dataset, self.patch_size, self.patch_size)
        if stage in ["test"]:
            self.test_sampler = GridGeoSampler(self.val_dataset, self.patch_size, self.patch_size)



# Copyright (c) IBM Corp. All rights reserved.
# Licensed under the MIT License.

import logging
from functools import partial
from pathlib import Path

import torch
from timm.models import FeatureInfo
from timm.models._builder import build_model_with_cfg
from timm.models._registry import generate_default_cfgs, register_model
from torch import nn

from terratorch.datasets import HLSBands
from terratorch.models.backbones.prithvi_select_patch_embed_weights import prithvi_select_patch_embed_weights
from terratorch.models.backbones.vit_encoder_decoder import TemporalViTEncoder

PRETRAINED_BANDS = [
    HLSBands.BLUE,
    HLSBands.GREEN,
    HLSBands.RED,
    HLSBands.NIR_NARROW,
    HLSBands.SWIR_1,
    HLSBands.SWIR_2,
]

def _cfg(file: Path = "", **kwargs) -> dict:
    return {
        "file": file,
        "source": "file",
        "input_size": (6, 224, 224),
        "license": "mit",
        # "first_conv": "patch_embed.proj",
        **kwargs,
    }

default_cfgs = generate_default_cfgs(
    {
        "prithvi_vit_100": {
            "hf_hub_id": "ibm-nasa-geospatial/Prithvi-100M",
            "hf_hub_filename": "Prithvi_100M.pt"
        }
    }
)

def checkpoint_filter_fn(
    state_dict, model: TemporalViTEncoder, pretrained_bands: list[HLSBands | int], model_bands: list[HLSBands | int]
) -> dict:
    if "pos_embed" in state_dict:
        del state_dict["pos_embed"]
    if "decoder_pos_embed" in state_dict:
        del state_dict["decoder_pos_embed"]

    if model.encoder_only:
        encoder_only_dict = {}
        for k, v in state_dict.items():
            if "decoder" in k:
                continue
            if "mask_token" in k:
                continue
            encoder_only_dict[k] = v
        state_dict = encoder_only_dict

    state_dict = prithvi_select_patch_embed_weights(state_dict, model, pretrained_bands, model_bands)

    return state_dict

def _create_prithvi(
    variant: str,
    pretrained: bool = False,  # noqa: FBT001, FBT002
    pretrained_bands: list[HLSBands] | None = None,
    model_bands: list[HLSBands | int] | None = None,
    **kwargs,
) -> TemporalViTEncoder:
    if pretrained_bands is None:
        pretrained_bands = PRETRAINED_BANDS

    # Little hack because VIT does not support timm's features_only
    # so we do it ourselves
    encoder_only = kwargs.get("features_only", False)
    if "features_only" in kwargs:
        kwargs = {k: v for k, v in kwargs.items() if k != "features_only"}

    kwargs["in_chans"] = len(model_bands)

    def checkpoint_filter_wrapper_fn(state_dict, model):
        return checkpoint_filter_fn(state_dict, model, pretrained_bands, model_bands)

    model = build_model_with_cfg(
        TemporalViTEncoder,
        variant,
        pretrained,
        pretrained_filter_fn=checkpoint_filter_wrapper_fn,
        pretrained_strict=True,
        encoder_only=encoder_only,
        **kwargs,
    )

    if encoder_only:
        default_out_indices = list(range(len(model.blocks)))
        out_indices = kwargs.get("out_indices", default_out_indices)
        if "out_indices" in kwargs:
            kwargs = {k: v for k, v in kwargs.items() if k != "out_indices"}
        model.feature_info = FeatureInfo(model.feature_info, out_indices)
        model.encode_decode_forward = model.forward
        model.forward = model.forward_features
        model.model_bands = model_bands
        model.pretrained_bands = pretrained_bands

    return model

def create_prithvi_vit_100(
    model_name: str,
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    **kwargs,
) -> TemporalViTEncoder:
    """Prithvi ViT 100M"""
    pretrained_bands = PRETRAINED_BANDS
    if bands is None:
        bands = pretrained_bands
        logging.info(
            f"Model bands not passed. Assuming bands are ordered in the same way as {PRETRAINED_BANDS}.\
            Pretrained patch_embed layer may be misaligned with current bands"
        )

    model_args = {
        "patch_size": 16,
        "embed_dim": 768,
        "depth": 12,
        "num_heads": 12,
        "decoder_embed_dim": 512,
        "decoder_depth": 8,
        "decoder_num_heads": 16,
        "mlp_ratio": 4,
        "norm_layer": partial(nn.LayerNorm, eps=1e-6),
        "num_frames": 1,
    }
    model = _create_prithvi(
        model_name,
        pretrained=pretrained,
        model_bands=bands,
        pretrained_bands=pretrained_bands,
        **dict(model_args, **kwargs),
    )
    return model


def create_prithvi_vit_300(
    model_name: str,
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    **kwargs,
) -> TemporalViTEncoder:
    """Prithvi ViT 300M"""
    pretrained_bands = PRETRAINED_BANDS
    if bands is None:
        bands = pretrained_bands
        logging.info(
            f"Model bands not passed. Assuming bands are ordered in the same way as {PRETRAINED_BANDS}.\
            Pretrained patch_embed layer may be misaligned with current bands"
        )
    model_args = {
        "patch_size": 16,
        "embed_dim": 1024,
        "depth": 24,
        "num_heads": 16,
        "decoder_embed_dim": 512,
        "decoder_depth": 8,
        "decoder_num_heads": 16,
        "mlp_ratio": 4,
        "norm_layer": partial(nn.LayerNorm, eps=1e-6),
        "num_frames": 1,
    }
    model = _create_prithvi(
        model_name,
        pretrained=pretrained,
        pretrained_bands=pretrained_bands,
        model_bands=bands,
        **dict(model_args, **kwargs),
    )
    return model


@register_model
def prithvi_vit_tiny(
    bands: list[HLSBands] | None = None,
    **kwargs,
) -> TemporalViTEncoder:
    """Prithvi ViT tiny"""
    pretrained_bands = PRETRAINED_BANDS
    if bands is None:
        bands = pretrained_bands
    model_args = {
        "patch_size": 16,
        "embed_dim": 256,
        "depth": 4,
        "num_heads": 4,
        "decoder_embed_dim": 128,
        "decoder_depth": 4,
        "decoder_num_heads": 4,
        "mlp_ratio": 4,
        "norm_layer": partial(nn.LayerNorm, eps=1e-6),
        "num_frames": 1,
        "model_bands": bands,
    }
    model = _create_prithvi("prithvi_vit_tiny", **dict(model_args, **kwargs))
    return model

@register_model
def prithvi_vit_100(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    **kwargs,
) -> TemporalViTEncoder:
    return create_prithvi_vit_100("prithvi_vit_100", pretrained, bands, **kwargs)


@register_model
def prithvi_vit_300(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    **kwargs,
) -> TemporalViTEncoder:
    return create_prithvi_vit_300("prithvi_vit_300", pretrained, bands, **kwargs)



from typing import Any
import torch

import albumentations as A
import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential
from terratorch.datasets import MBigEarthNonGeo
from terratorch.datamodules.utils import wrap_in_compose_is_list

MEANS = {
    "COASTAL_AEROSOL": 378.4027,  
    "BLUE": 482.2730,  
    "GREEN": 706.5345,  
    "RED": 720.9285, 
    "RED_EDGE_1": 1100.6688, 
    "RED_EDGE_2": 1909.2914,
    "RED_EDGE_3": 2191.6985, 
    "NIR_BROAD": 2336.8706, 
    "NIR_NARROW": 2394.7449, 
    "WATER_VAPOR": 2368.3127, 
    "SWIR_1": 1875.2487, 
    "SWIR_2": 1229.3818
}

STDS =  {
    "COASTAL_AEROSOL": 157.5666, 
    "BLUE": 255.0429, 
    "GREEN": 303.1750, 
    "RED": 391.2943, 
    "RED_EDGE_1": 380.7916, 
    "RED_EDGE_2": 551.6558, 
    "RED_EDGE_3": 638.8196,
    "NIR_BROAD": 744.2009,
    "NIR_NARROW": 675.4041, 
    "WATER_VAPOR": 561.0154, 
    "SWIR_1": 563.4095, 
    "SWIR_2": 479.1786
}

class MBigEarthNonGeoDataModule(NonGeoDataModule):
    def __init__(
        self, 
        batch_size: int = 8, 
        num_workers: int = 0, 
        data_root: str = "./",
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        **kwargs: Any
    ) -> None:

        super().__init__(MBigEarthNonGeo, batch_size, num_workers, **kwargs)
        
        bands = kwargs.get("bands", MBigEarthNonGeo.all_band_names)
        self.means = torch.tensor([MEANS[b] for b in bands])
        self.stds = torch.tensor([STDS[b] for b in bands])
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.data_root = data_root
        self.aug = AugmentationSequential(K.Normalize(self.means, self.stds), data_keys=["image"]) if aug is None else aug
    
    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(  
                split="train", data_root=self.data_root, transform=self.train_transform, **self.kwargs
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="val", data_root=self.data_root, transform=self.val_transform, **self.kwargs
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",data_root=self.data_root, transform=self.test_transform, **self.kwargs
            )



import pytest
import torch

from terratorch.models import PrithviModelFactory
from terratorch.models.backbones.prithvi_vit import PRETRAINED_BANDS
from terratorch.tasks import ClassificationTask, PixelwiseRegressionTask, SemanticSegmentationTask

NUM_CHANNELS = 6
NUM_CLASSES = 2
EXPECTED_SEGMENTATION_OUTPUT_SHAPE = (1, NUM_CLASSES, 224, 224)


@pytest.fixture(scope="session")
def model_factory() -> str:
    return "PrithviModelFactory"


@pytest.fixture(scope="session")
def model_input() -> torch.Tensor:
    return torch.ones((1, NUM_CHANNELS, 224, 224))


@pytest.mark.parametrize("backbone",["prithvi_vit_100", "prithvi_vit_300", "prithvi_swin_B"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
@pytest.mark.parametrize("loss", ["ce", "jaccard", "focal", "dice"])
def test_create_segmentation_task(backbone, decoder, loss, model_factory: PrithviModelFactory):
    SemanticSegmentationTask(
        {
            "backbone": backbone,
            "decoder": decoder,
            "in_channels": NUM_CHANNELS,
            "bands": PRETRAINED_BANDS,
            "pretrained": False,
            "num_classes": NUM_CLASSES,
        },
        model_factory,
        loss=loss
    )


@pytest.mark.parametrize("backbone", ["prithvi_vit_100", "prithvi_vit_300", "prithvi_swin_B"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
@pytest.mark.parametrize("loss", ["mae", "rmse", "huber"])
def test_create_regression_task(backbone, decoder, loss, model_factory: PrithviModelFactory):
    PixelwiseRegressionTask(
        {
            "backbone": backbone,
            "decoder": decoder,
            "in_channels": NUM_CHANNELS,
            "bands": PRETRAINED_BANDS,
            "pretrained": False,
        },
        model_factory,
        loss=loss
    )


@pytest.mark.parametrize("backbone", ["prithvi_vit_100", "prithvi_vit_300", "prithvi_swin_B"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
@pytest.mark.parametrize("loss", ["ce", "bce", "jaccard", "focal"])
def test_create_classification_task(backbone, decoder, loss, model_factory: PrithviModelFactory):
    ClassificationTask(
        {
            "backbone": backbone,
            "decoder": decoder,
            "in_channels": NUM_CHANNELS,
            "bands": PRETRAINED_BANDS,
            "pretrained": False,
            "num_classes": NUM_CLASSES,
        },
        model_factory,
        loss=loss
    )



import pytest
import timm
import torch

import terratorch  # noqa: F401

NUM_CHANNELS = 6
NUM_FRAMES = 3


@pytest.fixture
def input_224():
    return torch.ones((1, NUM_CHANNELS, 224, 224))


@pytest.fixture
def input_512():
    return torch.ones((1, NUM_CHANNELS, 512, 512))


@pytest.fixture
def input_224_multitemporal():
    return torch.ones((1, NUM_CHANNELS, NUM_FRAMES, 224, 224))


@pytest.fixture
def input_386():
    return torch.ones((1, NUM_CHANNELS, 386, 386))


@pytest.mark.parametrize("model_name", ["prithvi_vit_100", "prithvi_vit_300", "prithvi_swin_B"])
@pytest.mark.parametrize("test_input", ["input_224", "input_512"])
def test_can_create_backbones_from_timm(model_name, test_input, request):
    backbone = timm.create_model(model_name, pretrained=False)
    input_tensor = request.getfixturevalue(test_input)
    backbone(input_tensor)


@pytest.mark.parametrize("model_name", ["prithvi_vit_100", "prithvi_vit_300", "prithvi_swin_B"])
@pytest.mark.parametrize("test_input", ["input_224", "input_512"])
def test_can_create_backbones_from_timm_features_only(model_name, test_input, request):
    backbone = timm.create_model(model_name, pretrained=False, features_only=True)
    input_tensor = request.getfixturevalue(test_input)
    backbone(input_tensor)


@pytest.mark.parametrize("model_name", ["prithvi_vit_100", "prithvi_vit_300"])
def test_vit_models_accept_multitemporal(model_name, input_224_multitemporal):
    backbone = timm.create_model(model_name, pretrained=False, num_frames=NUM_FRAMES)
    backbone(input_224_multitemporal)


#def test_swin_models_accept_non_divisible_by_patch_size(input_386):
#    backbone = timm.create_model("prithvi_swin_90_us", pretrained=False, num_frames=NUM_FRAMES)
#    backbone(input_386)



from typing import Any, Iterable
import torch

import albumentations as A
import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential
from terratorch.datasets import MSo2SatNonGeo
from terratorch.datamodules.utils import wrap_in_compose_is_list

MEANS = {
    "VH_REAL": 0.00030114364926703274,
    "BLUE": 0.12951652705669403,
    "VH_IMAGINARY": -5.6475887504348066e-06,
    "GREEN": 0.11734361201524734,
    "VV_REAL": 0.000289927760604769,
    "RED": 0.11374464631080627,
    "VV_IMAGINARY": -0.0005758664919994771,
    "VH_LEE_FILTERED": 0.056090500205755234,
    "RED_EDGE_1": 0.12693354487419128,
    "VV_LEE_FILTERED": 0.33695536851882935,
    "RED_EDGE_2": 0.16917912662029266,
    "VH_LEE_FILTERED_REAL": -0.0026130808982998133,
    "RED_EDGE_3": 0.19080990552902222,
    "NIR_BROAD": 0.18381330370903015,
    "VV_LEE_FILTERED_IMAGINARY": -0.00021384705905802548,
    "NIR_NARROW": 0.20517952740192413,
    "SWIR_1": 0.1762811541557312,
    "SWIR_2": 0.1286638230085373
}

STDS =  {
    "VH_REAL": 0.20626230537891388,
    "BLUE": 0.040680479258298874,
    "VH_IMAGINARY": 0.19834314286708832,
    "GREEN": 0.05125178396701813,
    "VV_REAL": 0.5187134146690369,
    "RED": 0.07254913449287415,
    "VV_IMAGINARY": 0.519291877746582,
    "VH_LEE_FILTERED": 1.8058000802993774,
    "RED_EDGE_1": 0.06872648745775223,
    "VV_LEE_FILTERED": 4.893222808837891,
    "RED_EDGE_2": 0.07402216643095016,
    "VH_LEE_FILTERED_REAL": 0.8903943300247192,
    "RED_EDGE_3": 0.08412779122591019,
    "NIR_BROAD": 0.08534552156925201,
    "VV_LEE_FILTERED_IMAGINARY": 1.2786953449249268,
    "NIR_NARROW": 0.09248979389667511,
    "SWIR_1": 0.10270608961582184,
    "SWIR_2": 0.09284552931785583
}

class MSo2SatNonGeoDataModule(NonGeoDataModule):
    def __init__(
        self, 
        batch_size: int = 8, 
        num_workers: int = 0, 
        data_root: str = "./",
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        **kwargs: Any
    ) -> None:

        super().__init__(MSo2SatNonGeo, batch_size, num_workers, **kwargs)
        
        bands = kwargs.get("bands", MSo2SatNonGeo.all_band_names)
        self.means = torch.tensor([MEANS[b] for b in bands])
        self.stds = torch.tensor([STDS[b] for b in bands])
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.data_root = data_root
        self.aug = AugmentationSequential(K.Normalize(self.means, self.stds), data_keys=["image"]) if aug is None else aug
    
    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(  
                split="train", data_root=self.data_root, transform=self.train_transform, **self.kwargs
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="val", data_root=self.data_root, transform=self.val_transform, **self.kwargs
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",data_root=self.data_root, transform=self.test_transform, **self.kwargs
            )


# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# References:
# timm: https://github.com/rwightman/pytorch-image-models/tree/master/timm
# DeiT: https://github.com/facebookresearch/deit
# --------------------------------------------------------

from functools import lru_cache

import numpy as np
import torch
from einops import rearrange
from timm.layers import to_2tuple
from timm.models.vision_transformer import Block
from torch import Tensor, nn

B_C_H_W_SHAPE_LEN = 4


def get_1d_sincos_pos_embed_from_grid(embed_dim, pos):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    if embed_dim % 2 != 0:
        msg = "Embed dim must be divisible by 2"
        raise Exception(msg)
    omega = np.arange(embed_dim // 2, dtype=np.float32)
    omega /= embed_dim / 2.0
    omega = 1.0 / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = np.einsum("m,d->md", pos, omega)  # (M, D/2), outer product

    emb_sin = np.sin(out)  # (M, D/2)
    emb_cos = np.cos(out)  # (M, D/2)

    emb = np.concatenate([emb_sin, emb_cos], axis=1)  # (M, D)
    return emb


@lru_cache(maxsize=100, typed=False)
def get_3d_sincos_pos_embed(embed_dim: int, grid_size: tuple[int, int, int], cls_token: bool = False):
    # Copyright (c) Meta Platforms, Inc. and affiliates.
    # All rights reserved.

    # This source code is licensed under the license found in the
    # LICENSE file in the root directory of this source tree.
    # --------------------------------------------------------
    # Position embedding utils
    # --------------------------------------------------------
    """
    grid_size: 3d tuple of grid size: t, h, w
    return:
    pos_embed: L, D
    """

    if embed_dim % 16 != 0:
        msg = "Embed dim must be divisible by 16"
        raise Exception(msg)

    t_size, h_size, w_size = grid_size

    w_embed_dim = embed_dim // 16 * 6
    h_embed_dim = embed_dim // 16 * 6
    t_embed_dim = embed_dim // 16 * 4

    w_pos_embed = get_1d_sincos_pos_embed_from_grid(w_embed_dim, np.arange(w_size))
    h_pos_embed = get_1d_sincos_pos_embed_from_grid(h_embed_dim, np.arange(h_size))
    t_pos_embed = get_1d_sincos_pos_embed_from_grid(t_embed_dim, np.arange(t_size))

    w_pos_embed = np.tile(w_pos_embed, (t_size * h_size, 1))
    h_pos_embed = np.tile(np.repeat(h_pos_embed, w_size, axis=0), (t_size, 1))
    t_pos_embed = np.repeat(t_pos_embed, h_size * w_size, axis=0)

    pos_embed = np.concatenate((w_pos_embed, h_pos_embed, t_pos_embed), axis=1)

    if cls_token:
        pos_embed = np.concatenate([np.zeros([1, embed_dim]), pos_embed], axis=0)
    return pos_embed


class PatchEmbed(nn.Module):
    """Frames of 2D Images to Patch Embedding
    The 3D version of timm.models.vision_transformer.PatchEmbed
    """

    def __init__(
        self,
        img_size=224,
        patch_size=16,
        num_frames=3,
        tubelet_size=1,
        in_chans=3,
        embed_dim=768,
        norm_layer=None,
        flatten=True,  # noqa: FBT002
        bias=True,  # noqa: FBT002
    ):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_frames = num_frames
        self.tubelet_size = tubelet_size
        self.in_chans = in_chans
        self.grid_size = (
            num_frames // tubelet_size,
            img_size[0] // patch_size[0],
            img_size[1] // patch_size[1],
        )
        self.num_patches = self.grid_size[0] * self.grid_size[1] * self.grid_size[2]
        self.flatten = flatten

        self.proj = nn.Conv3d(
            in_chans,
            embed_dim,
            kernel_size=(tubelet_size, patch_size[0], patch_size[1]),
            stride=(tubelet_size, patch_size[0], patch_size[1]),
            bias=bias,
        )
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        if len(x.shape) == B_C_H_W_SHAPE_LEN and self.num_frames == 1:
            x = x.reshape(-1, self.in_chans, self.num_frames, *x.shape[-2:])
        B, C, T, H, W = x.shape  # noqa: N806
        x = self.proj(x)
        # Hp, Wp = x.shape[3], x.shape[4]
        if self.flatten:
            x = x.flatten(2).transpose(1, 2)  # B,C,T,H,W -> B,C,L -> B,L,C
        x = self.norm(x)
        return x


class TemporalViTEncoder(nn.Module):
    """Encoder from an ViT with capability to take in temporal input.

    This class defines an encoder taken from a ViT architecture.
    """

    # see https://github.com/huggingface/pytorch-image-models/blob/d5f1525334e1b111e4bfdf59fcd38eb9f8c9d3de/timm/models/vision_transformer.py#L407C15-L407C15
    def __init__(
        self,
        pretrain_img_size: int = 224,
        patch_size: int = 16,
        num_frames: int = 3,
        tubelet_size: int = 1,
        in_chans: int = 3,
        embed_dim: int = 1024,
        depth: int = 24,
        num_heads: int = 16,
        decoder_embed_dim: int = 512,
        decoder_depth: int = 8,
        decoder_num_heads: int = 16,
        mlp_ratio: float = 4.0,
        norm_layer: nn.Module = nn.LayerNorm,
        norm_pix_loss: bool = False,  # noqa: FBT001, FBT002
        encoder_only: bool = True,  # noqa: FBT001, FBT002
        **kwargs,  # timm parameters that may be passed  # noqa: ARG002
    ):
        """
        Args:
            pretrain_img_size (int, optional): Input image size for SSL. Ignored for forward_feature. Defaults to 224.
            patch_size (int, optional): Patch size to be used by the transformer. Defaults to 16.
            num_frames (int, optional): Number of frames (temporal dimension) to be input to the encoder. Defaults to 1.
            tubelet_size (int, optional): Tubelet size used in patch embedding. Defaults to 1.
            in_chans (int, optional): Number of input channels. Defaults to 3.
            embed_dim (int, optional): Embedding dimension. Defaults to 1024.
            depth (int, optional): Encoder depth. Defaults to 24.
            num_heads (int, optional): Number of heads used in the encoder blocks. Defaults to 16.
            mlp_ratio (float, optional): Ratio to be used for the size of the MLP in encoder blocks. Defaults to 4.0.
            norm_layer (nn.Module, optional): Norm layer to be used. Defaults to nn.LayerNorm.
            norm_pix_loss (bool, optional): Whether to use Norm Pix Loss. Defaults to False.
            pretrained (str, optional): Path to pretrained encoder weights. Defaults to None.
        """
        super().__init__()

        # --------------------------------------------------------------------------
        # MAE encoder specifics
        self.patch_embed = PatchEmbed(pretrain_img_size, patch_size, num_frames, tubelet_size, in_chans, embed_dim)
        self.feature_info = []
        self.in_chans = in_chans
        self.num_frames = num_frames
        self.embed_dim = embed_dim
        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.encoder_only = encoder_only
        # initialize (and freeze) pos_embed by sin-cos embedding
        pos_embed = get_3d_sincos_pos_embed(self.embed_dim, self.patch_embed.grid_size, cls_token=True)
        self.register_buffer(
            "pos_embed",
            torch.from_numpy(pos_embed).float().unsqueeze(0),
            persistent=False,
        )

        self.blocks = [
            Block(embed_dim, num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer) for i in range(depth)
        ]
        for i, _ in enumerate(self.blocks):
            self.feature_info.append({"num_chs": embed_dim * num_frames, "reduction": 1, "module": f"blocks.{i}"})
        self.blocks = nn.ModuleList(self.blocks)
        self.norm = norm_layer(embed_dim)
        # --------------------------------------------------------------------------

        # --------------------------------------------------------------------------
        # MAE decoder specifics
        if not encoder_only:
            self.decoder_embed_dim = decoder_embed_dim
            self.decoder_embed = nn.Linear(embed_dim, decoder_embed_dim, bias=True)

            self.mask_token = nn.Parameter(torch.zeros(1, 1, decoder_embed_dim))

            # self.decoder_pos_embed = torch.zeros(
            #     1, num_patches + 1, decoder_embed_dim
            # )  # fixed sin-cos embedding

            self.decoder_blocks = nn.ModuleList(
                [
                    Block(
                        decoder_embed_dim,
                        decoder_num_heads,
                        mlp_ratio,
                        qkv_bias=True,
                        norm_layer=norm_layer,
                    )
                    for i in range(decoder_depth)
                ]
            )

            # decoder_pos_embed = get_3d_sincos_pos_embed(
            #     self.decoder_embed_dim, self.patch_embed.grid_size, cls_token=True
            # )
            # self.register_buffer(
            #     "decoder_pos_embed",
            #     torch.from_numpy(decoder_pos_embed).float().unsqueeze(0),
            #     persistent=False,
            # )

            self.decoder_norm = norm_layer(decoder_embed_dim)
            self.decoder_pred = nn.Linear(
                decoder_embed_dim,
                tubelet_size * patch_size * patch_size * in_chans,
                bias=True,
            )  # decoder to patch
        # --------------------------------------------------------------------------

        self.norm_pix_loss = norm_pix_loss

        self.initialize_weights()

        self.embed_dims = [self.embed_dim] * len(self.blocks)

    def initialize_weights(self):
        # initialization

        # initialize patch_embed like nn.Linear (instead of nn.Conv2d)
        w = self.patch_embed.proj.weight.data
        torch.nn.init.xavier_uniform_(w.view([w.shape[0], -1]))

        # timm's trunc_normal_(std=.02) is effectively normal_(std=0.02) as cutoff is too big (2.)
        torch.nn.init.normal_(self.cls_token, std=0.02)
        if not self.encoder_only:
            torch.nn.init.normal_(self.mask_token, std=0.02)

        # initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def patchify(self, imgs: torch.Tensor) -> torch.Tensor:
        """
        imgs: B, C, T, H, W
        x: B, L, D
        """
        p = self.patch_embed.patch_size[0]
        tub = self.patch_embed.tubelet_size
        x = rearrange(imgs, "b c (t tub) (h p) (w q) -> b (t h w) (tub p q c)", tub=tub, p=p, q=p)

        return x

    def unpatchify(self, x: torch.Tensor) -> torch.Tensor:
        """
        x: B, L, D
        imgs: B, C, T, H, W
        """
        p = self.patch_embed.patch_size[0]
        num_p = self.patch_embed.img_size[0] // p
        tub = self.patch_embed.tubelet_size
        imgs = rearrange(
            x,
            "b (t h w) (tub p q c) -> b c (t tub) (h p) (w q)",
            h=num_p,
            w=num_p,
            tub=tub,
            p=p,
            q=p,
        )
        return imgs

    def random_masking(self, x: torch.Tensor, mask_ratio: float) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Perform per-sample random masking by per-sample shuffling.
        Per-sample shuffling is done by argsort random noise.
        x: [N, L, D], sequence
        """
        N, L, D = x.shape  # batch, length, dim  # noqa: N806
        len_keep = int(L * (1 - mask_ratio))

        noise = torch.rand(N, L, device=x.device)  # noise in [0, 1]

        # sort noise for each sample
        ids_shuffle = torch.argsort(noise, dim=1)  # ascend: small is keep, large is remove
        ids_restore = torch.argsort(ids_shuffle, dim=1)

        # keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]
        x_masked = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, D))

        # generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([N, L], device=x.device)
        mask[:, :len_keep] = 0
        # unshuffle to get the binary mask
        mask = torch.gather(mask, dim=1, index=ids_restore)

        return x_masked, mask, ids_restore

    def forward_encoder(
        self, x: torch.Tensor, mask_ratio: float = 0.0
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        # embed patches
        if len(x.shape) == B_C_H_W_SHAPE_LEN and self.patch_embed.num_frames == 1:
            x = x.reshape(-1, self.in_chans, 1, *x.shape[-2:])
        t, h, w = x.shape[-3:]
        x = self.patch_embed(x)
        pos_embed = torch.from_numpy(get_3d_sincos_pos_embed(self.embed_dim, (t, h // 16, w // 16), cls_token=True)).to(
            x
        )
        # add pos embed w/o cls token
        x = x + pos_embed[1:, :]

        # masking: length -> length * mask_ratio
        x, mask, ids_restore = self.random_masking(x, mask_ratio)

        # append cls token
        cls_token = self.cls_token + pos_embed[:1, :]
        cls_tokens = cls_token.expand(x.shape[0], -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        # apply Transformer blocks
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)

        return x, mask, ids_restore, (t, h, w)

    def forward_decoder(self, x: torch.Tensor, ids_restore: torch.Tensor, dim_info: tuple) -> torch.Tensor:
        if self.encoder_only:
            msg = (
                "Cannot run forward decoder method with self.encoder_only. Pass encoder_only=False to use the decoder."
            )
            raise Exception(msg)
        # embed tokens
        x = self.decoder_embed(x)
        t, h, w = dim_info
        decoder_pos_embed = torch.from_numpy(
            get_3d_sincos_pos_embed(self.decoder_embed_dim, (t, h // 16, w // 16), cls_token=True)
        ).to(x)

        # append mask tokens to sequence
        mask_tokens = self.mask_token.repeat(x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        x_ = torch.cat([x[:, 1:, :], mask_tokens], dim=1)  # no cls token
        x_ = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle
        x = torch.cat([x[:, :1, :], x_], dim=1)  # append cls token

        # add pos embed
        x = x + decoder_pos_embed

        # apply Transformer blocks
        for blk in self.decoder_blocks:
            x = blk(x)
        x = self.decoder_norm(x)

        # predictor projection
        x = self.decoder_pred(x)

        # remove cls token
        x = x[:, 1:, :]

        return x

    def forward_loss(self, imgs, pred, mask) -> torch.Tensor:
        """
        imgs: B, C, T, H, W
        target: B, L, D
        pred: B, L, D
        mask: B, L. 0 is keep, 1 is remove,
        """
        if len(imgs.shape) == B_C_H_W_SHAPE_LEN and self.num_frames == 1:
            imgs = imgs.reshape(-1, self.in_chans, self.num_frames, *imgs.shape[-2:])
        target = self.patchify(imgs)
        if self.norm_pix_loss:
            mean = target.mean(dim=-1, keepdim=True)
            var = target.var(dim=-1, keepdim=True)
            target = (target - mean) / (var + 1.0e-6) ** 0.5

        loss = (pred - target) ** 2
        loss = loss.mean(dim=-1)  # [N, L], mean loss per patch

        loss = (loss * mask).sum() / mask.sum()  # mean loss on removed patches
        return loss

    def forward(self, imgs, mask_ratio=0.0) -> torch.Tensor:
        latent, mask, ids_restore, dim_info = self.forward_encoder(imgs, mask_ratio)
        pred = self.forward_decoder(latent, ids_restore, dim_info)
        _ = self.forward_loss(imgs, pred, mask)
        return pred

    def forward_features(self, x) -> list[torch.Tensor]:
        if len(x.shape) == B_C_H_W_SHAPE_LEN and self.patch_embed.num_frames == 1:
            x = x.reshape(-1, self.in_chans, 1, *x.shape[-2:])
        t, h, w = x.shape[-3:]
        # embed patches
        x = self.patch_embed(x)
        pos_embed = torch.from_numpy(get_3d_sincos_pos_embed(self.embed_dim, (t, h // 16, w // 16), cls_token=True)).to(
            x
        )
        # add pos embed w/o cls token
        x = x + pos_embed[1:, :]

        # append cls token
        cls_token = self.cls_token + pos_embed[:1, :]
        cls_tokens = cls_token.expand(x.shape[0], -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        # apply Transformer blocks
        out = []
        for blk in self.blocks:
            x = blk(x)
            out.append(x.clone())

        x = self.norm(x)
        out[-1] = x
        return out

    def prepare_features_for_image_model(self, features: list[Tensor]) -> list[Tensor]:
        out = []
        for x in features:
            x_no_token = x[:, 1:, :]
            # reshape and permute to channels first
            # encoded = rearrange(
            #     x_no_token,
            #     "batch (t h w) e -> batch (t e) h w",
            #     e=self.embed_dim,
            #     t=self.num_frames,
            #     h=int(np.sqrt(x_no_token.shape[1] // self.num_frames)),
            # )
            encoded = x_no_token.permute(0, 2, 1).reshape(
                x.shape[0],
                -1,
                int(np.sqrt(x_no_token.shape[1] // self.num_frames)),
                int(np.sqrt(x_no_token.shape[1] // self.num_frames)),
            )
            # encoded = (
            #     x_no_token.reshape(
            #         -1,
            #         int(np.sqrt(x_no_token.shape[1])),
            #         int(np.sqrt(x_no_token.shape[1])),
            #         self.embed_dim * self.num_frames,
            #     )
            #     .permute(0, 3, 1, 2)
            #     .contiguous()
            # )
            out.append(encoded)
        return out



import logging  # noqa: I001
import os
import warnings
from datetime import timedelta
from pathlib import Path
from typing import Any, Optional
import tempfile
import itertools
import yaml
import json
from jsonargparse import set_dumper
from copy import deepcopy

import albumentations
import cv2  # noqa: F401
import numpy as np
import rasterio
import torch

# Allows classes to be referenced using only the class name
import torchgeo.datamodules
from albumentations.pytorch import ToTensorV2  # noqa: F401
from lightning.fabric.utilities.types import _PATH  # noqa: F401
from lightning.pytorch import LightningDataModule, LightningModule, Trainer  # noqa: F401
from lightning.pytorch.callbacks import BasePredictionWriter, ModelCheckpoint, RichProgressBar
from lightning.pytorch.cli import ArgsType, LightningArgumentParser, LightningCLI, SaveConfigCallback  # noqa: F401
from torchgeo.trainers import BaseTask

from lightning.fabric.utilities.cloud_io import get_filesystem

import terratorch.datamodules
from terratorch.datasets.utils import HLSBands
import terratorch.tasks  # noqa: F401
from terratorch.datamodules import (  # noqa: F401
    GenericNonGeoClassificationDataModule,
    GenericNonGeoPixelwiseRegressionDataModule,
    GenericNonGeoSegmentationDataModule,
)
from terratorch.datasets.transforms import (
    FlattenTemporalIntoChannels,  # noqa: F401
    Rearrange,  # noqa: F401
    SelectBands,  # noqa: F401
    UnflattenTemporalFromChannels,  # noqa: F401
)

# GenericNonGeoRegressionDataModule,
from terratorch.models import PrithviModelFactory  # noqa: F401
from terratorch.models.model import AuxiliaryHead  # noqa: F401
from terratorch.tasks import (
    ClassificationTask,  # noqa: F401
    PixelwiseRegressionTask,  # noqa: F401
    SemanticSegmentationTask,  # noqa: F401
)


def flatten(list_of_lists):
    return list(itertools.chain.from_iterable(list_of_lists))


def open_tiff(fname):
    with rasterio.open(fname, "r") as src:
        data = src.read()
        metadata = src.meta
    return data, metadata


def is_one_band(img):
    return len(img.shape) == 2  # noqa: PLR2004


def write_tiff(img_wrt, filename, metadata):
    with rasterio.open(filename, "w", **metadata) as dest:
        if is_one_band(img_wrt):
            img_wrt = img_wrt[None]

        for i in range(img_wrt.shape[0]):
            dest.write(img_wrt[i, :, :], i + 1)
    return filename


def save_prediction(prediction, input_file_name, out_dir, dtype:str="int16"):
    mask, metadata = open_tiff(input_file_name)
    mask = np.where(mask == metadata["nodata"], 1, 0)
    mask = np.max(mask, axis=0)
    result = np.where(mask == 1, -1, prediction)

    ##### Save file to disk
    metadata["count"] = 1
    metadata["dtype"] = dtype
    metadata["compress"] = "lzw"
    metadata["nodata"] = -1
    file_name = os.path.basename(input_file_name)
    file_name_no_ext = os.path.splitext(file_name)[0]
    out_file_name = file_name_no_ext + "_pred.tif"
    logging.info(f"Saving output to {out_file_name} ...")
    write_tiff(result, os.path.join(out_dir, out_file_name), metadata)


class CustomWriter(BasePredictionWriter):
    """Callback class to write geospatial data to file."""

    def __init__(self, output_dir: str | None = None, write_interval: str = "epoch"):

        super().__init__(write_interval)

        self.output_dir = output_dir

    def write_on_epoch_end(self, trainer, pl_module, predictions, batch_indices):  # noqa: ARG002
        # this will create N (num processes) files in `output_dir` each containing
        # the predictions of it's respective rank

        # by default take self.output_dir. If None, look for one in trainer
        if self.output_dir is None:
            try:
                output_dir = trainer.predict_output_dir
            except AttributeError as err:
                msg = "Output directory must be passed to CustomWriter constructor or the `predict_output_dir`\
                attribute must be present in the trainer."
                raise Exception(msg) from err
        else:
            output_dir = self.output_dir

        if not os.path.exists(output_dir):
            os.mkdir(output_dir)

        for pred_batch, filename_batch in predictions:
            for prediction, file_name in zip(torch.unbind(pred_batch, dim=0), filename_batch, strict=False):
                save_prediction(prediction, file_name, output_dir, dtype=trainer.out_dtype)


def clean_config_for_deployment_and_dump(config: dict[str, Any]):
    deploy_config = deepcopy(config)
    ## General
    # drop ckpt_path
    deploy_config.pop("ckpt_path", None)
    # drop checkpoints
    deploy_config.pop("ModelCheckpoint", None)
    deploy_config.pop("StateDictModelCheckpoint", None)
    # drop optimizer and lr sheduler
    deploy_config.pop("optimizer", None)
    deploy_config.pop("lr_scheduler", None)
    ## Trainer
    # remove logging
    deploy_config["trainer"]["logger"] = False
    # remove callbacks
    deploy_config["trainer"].pop("callbacks", None)
    # remove default_root_dir
    deploy_config["trainer"].pop("default_root_dir", None)
    # set mixed precision by default for inference
    deploy_config["trainer"]["precision"] = "16-mixed"
    ## Model
    # set pretrained to false
    if "model_args" in deploy_config["model"]["init_args"]:
        deploy_config["model"]["init_args"]["model_args"]["pretrained"] = False

    return yaml.safe_dump(deploy_config)


class StudioDeploySaveConfigCallback(SaveConfigCallback):
    """Subclass of the SaveConfigCallback to save configs.
    Besides saving the usual config, will also save a modified config ready to be uploaded.
    Handles things like disabling the pretrained flag, removing loggers, ...
    """

    def __init__(
        self,
        parser,
        config,
        config_filename: str = "config.yaml",
        overwrite: bool = False,
        multifile: bool = False,
        save_to_log_dir: bool = True,
    ):
        super().__init__(parser, config, config_filename, overwrite, multifile, save_to_log_dir)
        set_dumper("deploy_config", clean_config_for_deployment_and_dump)

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if self.already_saved:
            return

        if self.save_to_log_dir:
            log_dir = trainer.log_dir or trainer.default_root_dir  # this broadcasts the directory
            if log_dir is None:
                msg = "log_dir was none. Please set a logger or set default_root_dir for the trainer"
                raise Exception(msg)
            config_path = os.path.join(log_dir, self.config_filename)
            fs = get_filesystem(log_dir)

            if not self.overwrite:
                # check if the file exists on rank 0
                file_exists = fs.isfile(config_path) if trainer.is_global_zero else False
                # broadcast whether to fail to all ranks
                file_exists = trainer.strategy.broadcast(file_exists)
                if file_exists:
                    msg = (
                        f"{self.__class__.__name__} expected {config_path} to NOT exist. Aborting to avoid overwriting"
                        " results of a previous run. You can delete the previous config file,"
                        " set `LightningCLI(save_config_callback=None)` to disable config saving,"
                        ' or set `LightningCLI(save_config_kwargs={"overwrite": True})` to overwrite the config file.'
                    )
                    raise RuntimeError(msg)

            if trainer.is_global_zero:
                # save only on rank zero to avoid race conditions.
                # the `log_dir` needs to be created as we rely on the logger to do it usually
                # but it hasn't logged anything at this point
                fs.makedirs(log_dir, exist_ok=True)
                self.parser.save(
                    self.config, config_path, skip_none=True, overwrite=self.overwrite, multifile=self.multifile
                )

        if trainer.is_global_zero:
            # also save the config that will be deployed
            config_name, config_ext = os.path.splitext(self.config_filename)
            config_name += "_deploy"
            config_name += config_ext
            config_path = os.path.join(log_dir, config_name)
            self.parser.save(
                self.config,
                config_path,
                format="deploy_config",
                skip_none=True,
                overwrite=self.overwrite,
                multifile=self.multifile,
            )
            self.already_saved = True

        # broadcast so that all ranks are in sync on future calls to .setup()
        self.already_saved = trainer.strategy.broadcast(self.already_saved)


class StateDictAwareModelCheckpoint(ModelCheckpoint):
    # necessary as we wish to have one model checkpoint with only state dict and one with standard lightning checkpoints
    # for this, the state key needs to be different, and thus to include this save_weights_only parameter
    def __init__(
        self,
        dirpath: _PATH | None = None,
        filename: str | None = None,
        monitor: str | None = None,
        verbose: bool = False,
        save_last: bool | None = None,
        save_top_k: int = 1,
        mode: str = "min",
        save_weights_only: bool = False,
        auto_insert_metric_name: bool = True,
        every_n_train_steps: int | None = None,
        train_time_interval: timedelta | None = None,
        every_n_epochs: int | None = None,
        save_on_train_epoch_end: bool | None = None,
        enable_version_counter: bool = True,
    ):
        super().__init__(
            dirpath,
            filename,
            monitor,
            verbose,
            save_last,
            save_top_k,
            save_weights_only,
            mode,
            auto_insert_metric_name,
            every_n_train_steps,
            train_time_interval,
            every_n_epochs,
            save_on_train_epoch_end,
            enable_version_counter,
        )

    @property
    def state_key(self) -> str:
        return self._generate_state_key(
            monitor=self.monitor,
            mode=self.mode,
            every_n_train_steps=self._every_n_train_steps,
            every_n_epochs=self._every_n_epochs,
            train_time_interval=self._train_time_interval,
            save_weights_only=self.save_weights_only,
        )


class MyLightningCLI(LightningCLI):
    def add_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
        parser.add_argument("--predict_output_dir", default=None)
        parser.add_argument("--out_dtype", default="int16")

        # parser.set_defaults({"trainer.enable_checkpointing": False})

        parser.add_lightning_class_args(StateDictAwareModelCheckpoint, "ModelCheckpoint")
        parser.set_defaults({"ModelCheckpoint.filename": "{epoch}", "ModelCheckpoint.monitor": "val/loss"})

        parser.add_lightning_class_args(StateDictAwareModelCheckpoint, "StateDictModelCheckpoint")
        parser.set_defaults(
            {
                "StateDictModelCheckpoint.filename": "{epoch}_state_dict",
                "StateDictModelCheckpoint.save_weights_only": True,
                "StateDictModelCheckpoint.monitor": "val/loss",
            }
        )

        parser.link_arguments("ModelCheckpoint.dirpath", "StateDictModelCheckpoint.dirpath")

    def instantiate_classes(self) -> None:
        super().instantiate_classes()
        # get the predict_output_dir. Depending on the value of run, it may be in the subcommand
        try:
            config = self.config.predict
        except AttributeError:
            config = self.config
        if hasattr(config, "predict_output_dir"):
            self.trainer.predict_output_dir = config.predict_output_dir
        
        if hasattr(config, "out_dtype"):
            self.trainer.out_dtype = config.out_dtype

def build_lightning_cli(
    args: ArgsType = None,
    run=True,  # noqa: FBT002
) -> LightningCLI:
    """Command-line interface to GeospatialCV."""
    # Taken from https://github.com/pangeo-data/cog-best-practices
    rasterio_best_practices = {
        "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
        "AWS_NO_SIGN_REQUEST": "YES",
        "GDAL_MAX_RAW_BLOCK_CACHE_SIZE": "200000000",
        "GDAL_SWATH_SIZE": "200000000",
        "VSI_CURL_CACHE_SIZE": "200000000",
    }
    os.environ.update(rasterio_best_practices)
    if os.environ.get("terratorch_FLOAT_32_PRECISION") is not None:
        precision = os.environ.get("terratorch_FLOAT_32_PRECISION")
        allowed_values = ["highest", "high", "medium"]
        if precision in allowed_values:
            torch.set_float32_matmul_precision(precision)
        else:
            warnings.warn(
                "Found terratorch_FLOAT_32_PRECISION env variable but value was set to precision.\
                Set to one of {allowed_values}. Will be ignored this run.",
                UserWarning,
                stacklevel=1,
            )

    return MyLightningCLI(
        model_class=BaseTask,
        subclass_mode_model=True,
        subclass_mode_data=True,
        seed_everything_default=0,
        save_config_callback=StudioDeploySaveConfigCallback if run else None,
        save_config_kwargs={"overwrite": True},
        args=args,
        # save only state_dict as well as full state. Only state_dict will be used for exporting the model
        trainer_defaults={"callbacks": [CustomWriter(write_interval="epoch")]},
        run=run,
    )


class LightningInferenceModel:
    """This class provides all the same functionalities as loading a model from the CLI,
    but makes the model accessible programatically.

    Example usage:
        model = LightningInferenceModel.from_config(
        <config path>,
        <checkpoint path>,
        <output dir path>,
        )
        model.inference(<path to inference data dir>)
    """

    def __init__(
        self,
        trainer: Trainer,
        model: LightningModule,
        datamodule: LightningDataModule,
        checkpoint_path: Path | None = None,
    ):
        self.trainer = trainer
        self.model = model
        self.datamodule = datamodule
        if checkpoint_path:
            weights = torch.load(checkpoint_path, map_location="cpu", weights_only=True)
            if "state_dict" in weights:
                weights = weights["state_dict"]
            weights = {k.replace("model.", ""): v for k, v in weights.items() if k.startswith("model.")}
            self.model.model.load_state_dict(weights)

        # dont write
        non_writing_callbacks = []
        for callback in self.trainer.callbacks:
            if isinstance(callback, BasePredictionWriter):
                continue
            non_writing_callbacks.append(callback)
        self.trainer.callbacks = non_writing_callbacks

    @staticmethod
    def from_config(
        config_path: Path,
        checkpoint_path: Path | None = None,
        predict_dataset_bands: list[str] | None = None,
    ):
        """
        Args:
            config_path (Path): Path to the config of the model to be loaded.
            checkpoint_path (Path): Path to the checkpoint to be loaded.
            predict_dataset_bands (list[str] | None, optional): List of bands present in input data.
                Defaults to None.
        """
        # use cli only to load
        arguments = [
            "--config",
            config_path,
        ]

        if predict_dataset_bands is not None:
            arguments.extend([ "--data.init_args.predict_dataset_bands",
            "[" + ",".join(predict_dataset_bands) + "]",])

        cli = build_lightning_cli(arguments, run=False)
        trainer = cli.trainer
        # disable logging metrics
        trainer.logger = None
        datamodule = cli.datamodule
        model = cli.model
        return LightningInferenceModel(trainer, model, datamodule, checkpoint_path=checkpoint_path)

    @staticmethod
    def from_task_and_datamodule(
        task: LightningModule,
        datamodule: LightningDataModule,
        checkpoint_path: Path | None = None,
    ):
        trainer = Trainer(
            # precision="16-mixed",
            callbacks=[RichProgressBar()],
        )
        return LightningInferenceModel(trainer, task, datamodule, checkpoint_path=checkpoint_path)

    def inference_on_dir(self, data_root: Path | None = None) -> tuple[torch.Tensor, list[str]]:
        """Perform inference on the given data root directory

        Args:
            data_root (Path): Directory with data to perform inference on
        Returns:
            A tuple with a torch tensor with all predictions and a list of corresponding input file paths

        """
        if data_root:
            self.datamodule.predict_root = data_root
        predictions = self.trainer.predict(model=self.model, datamodule=self.datamodule, return_predictions=True)
        concat_predictions = torch.cat([batch[0] for batch in predictions])
        concat_file_names = flatten([batch[1] for batch in predictions])

        return concat_predictions, concat_file_names

    def inference(self, file_path: Path) -> torch.Tensor:
        """Perform inference on a single input file path

        Args:
            file_path (Path): Path to a single input file.
        Returns:
            A torch tensor with the prediction
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            os.symlink(file_path, os.path.join(tmpdir, os.path.basename(file_path)))
            prediction, file_name = self.inference_on_dir(
                tmpdir,
            )
            return prediction.squeeze(0)



from terratorch.tasks.classification_tasks import ClassificationTask
from terratorch.tasks.regression_tasks import PixelwiseRegressionTask
from terratorch.tasks.segmentation_tasks import SemanticSegmentationTask
from terratorch.tasks.multilabel_classification_tasks import MultiLabelClassificationTask

__all__ = (
    "SemanticSegmentationTask",
    "PixelwiseRegressionTask",
    "ClassificationTask",
    "MultiLabelClassificationTask"
    "BATCH_IDX_FOR_VALIDATION_PLOTTING",
)



from typing import Any
import torch

import albumentations as A
import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential
from terratorch.datasets import MBeninSmallHolderCashewsNonGeo
from terratorch.datamodules.utils import wrap_in_compose_is_list


MEANS = {
    "COASTAL_AEROSOL": 520.1185302734375,
    "BLUE": 634.7583618164062,
    "GREEN": 892.461181640625,
    "RED": 880.7075805664062,
    "RED_EDGE_1": 1380.6409912109375,
    "RED_EDGE_2": 2233.432373046875,
    "RED_EDGE_3": 2549.379638671875,
    "NIR_BROAD": 2643.248046875,
    "NIR_NARROW": 2643.531982421875,
    "WATER_VAPOR": 2852.87451171875,
    "SWIR_1": 2463.933349609375,
    "SWIR_2": 1600.9207763671875,
    "CLOUD_PROBABILITY": 0.010281000286340714
}

STDS = {
    "COASTAL_AEROSOL": 204.2023468017578,
    "BLUE": 227.25344848632812,
    "GREEN": 222.32545471191406,
    "RED": 350.47235107421875,
    "RED_EDGE_1": 280.6436767578125,
    "RED_EDGE_2": 373.7521057128906,
    "RED_EDGE_3": 449.9236145019531,
    "NIR_BROAD": 414.6498107910156,
    "NIR_NARROW": 415.1019592285156,
    "WATER_VAPOR": 413.8980407714844,
    "SWIR_1": 494.97430419921875,
    "SWIR_2": 514.4229736328125,
    "CLOUD_PROBABILITY": 0.3447800576686859
}

class MBeninSmallHolderCashewsNonGeoDataModule(NonGeoDataModule):

    def __init__(
        self, 
        batch_size: int = 8, 
        num_workers: int = 0, 
        data_root: str = "./",
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        **kwargs: Any
    ) -> None:

        super().__init__(MBeninSmallHolderCashewsNonGeo, batch_size, num_workers, **kwargs)
        
        bands = kwargs.get("bands", MBeninSmallHolderCashewsNonGeo.all_band_names)
        self.means = torch.tensor([MEANS[b] for b in bands])
        self.stds = torch.tensor([STDS[b] for b in bands])
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.data_root = data_root
        self.aug = AugmentationSequential(K.Normalize(self.means, self.stds), data_keys=["image", "mask"]) if aug is None else aug
    
    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(  
                split="train", data_root=self.data_root, transform=self.train_transform, **self.kwargs
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="val", data_root=self.data_root, transform=self.val_transform, **self.kwargs
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",data_root=self.data_root, transform=self.test_transform, **self.kwargs
            )


from terratorch.models.prithvi_model_factory import PrithviModelFactory
from terratorch.models.satmae_model_factory import SatMAEModelFactory
from terratorch.models.scalemae_model_factory import ScaleMAEModelFactory
from terratorch.models.smp_model_factory import SMPModelFactory
from terratorch.models.timm_model_factory import TimmModelFactory

__all__ = (
    "PrithviModelFactory",
    "ClayModelFactory",
    "SatMAEModelFactory",
    "ScaleMAEModelFactory",
    "SMPModelFactory",
    "TimmModelFactory",
    "AuxiliaryHead",
    "AuxiliaryHeadWithDecoderWithoutInstantiatedHead",
)



from albumentations import BasicTransform, Compose, ImageOnlyTransform
from einops import rearrange
from torch import Tensor

N_DIMS_FOR_TEMPORAL = 4
N_DIMS_FLATTENED_TEMPORAL = 3


def albumentations_to_callable_with_dict(albumentation: list[BasicTransform] | None = None):
    if albumentation is None:
        return lambda x: x
    albumentation = Compose(albumentation)

    def fn(data):
        return albumentation(**data)

    return fn


class FlattenTemporalIntoChannels(ImageOnlyTransform):
    """Flatten the temporal dimension into channels"""

    def __init__(self):
        super().__init__(True, 1)

    def apply(self, img, **params):
        if len(img.shape) != N_DIMS_FOR_TEMPORAL:
            msg = f"Expected input temporal sequence to have {N_DIMS_FOR_TEMPORAL} dimensions, but got {len(img.shape)}"
            raise Exception(msg)
        rearranged = rearrange(img, "time height width channels -> height width (time channels)")
        return rearranged

    def get_transform_init_args_names(self):
        return ()


class UnflattenTemporalFromChannels(ImageOnlyTransform):
    """Flatten the temporal dimension into channels
    Assumes channels first (usually should be run after ToTensorV2)"""

    def __init__(self, n_timesteps: int | None = None, n_channels: int | None = None):
        super().__init__(True, 1)
        if n_timesteps is None and n_channels is None:
            msg = "One of n_timesteps or n_channels must be provided"
            raise Exception(msg)
        self.additional_info = {"channels": n_channels} if n_channels else {"time": n_timesteps}

    def apply(self, img, **params):
        if len(img.shape) != N_DIMS_FLATTENED_TEMPORAL:
            msg = f"Expected input temporal sequence to have {N_DIMS_FLATTENED_TEMPORAL} dimensions\
                , but got {len(img.shape)}"
            raise Exception(msg)

        rearranged = rearrange(
            img, "(time channels) height width -> channels time height width", **self.additional_info
        )
        return rearranged

    def get_transform_init_args_names(self):
        return ("n_timesteps", "n_channels")


class Rearrange(ImageOnlyTransform):
    """Flatten the temporal dimension into channels"""

    def __init__(
        self, rearrange: str, rearrange_args: dict[str, int] | None = None, always_apply: bool = True, p: float = 1
    ):
        super().__init__(always_apply, p)
        self.rearrange = rearrange
        self.vars = rearrange_args if rearrange_args else {}

    def apply(self, img, **params):
        return rearrange(img, self.rearrange, **self.vars)

    def get_transform_init_args_names(self):
        return "rearrange"


class SelectBands(ImageOnlyTransform):
    """Select a subset of the input bands"""

    def __init__(self, band_indices: list[int]):
        super().__init__(True, 1)
        self.band_indices = band_indices

    def apply(self, img, **params):
        return img[..., self.band_indices]

    def get_transform_init_args_names(self):
        return "band_indices"



import dataclasses
import glob
import os
from pathlib import Path
from typing import Any

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import rioxarray
from matplotlib import cm
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from torch import Tensor
from torchgeo.datasets import NonGeoDataset, RasterDataset


class FireScarsNonGeo(NonGeoDataset):
    """NonGeo dataset implementation for fire scars."""

    def __init__(self, data_root: Path) -> None:
        super().__init__()
        self.image_files = sorted(glob.glob(os.path.join(data_root, "subsetted*_merged.tif")))
        self.segmentation_mask_files = sorted(glob.glob(os.path.join(data_root, "subsetted*.mask.tif")))
        self.rgb_indices = [0, 1, 2]

    def __len__(self) -> int:
        return len(self.image_files)

    def __getitem__(self, index: int) -> dict[str, Any]:
        output = {
            "image": self._load_file(self.image_files[index]).astype(np.float32),
            "mask": self._load_file(self.segmentation_mask_files[index]).astype(np.int64),
        }
        return output

    def _load_file(self, path: Path):
        data = rioxarray.open_rasterio(path)
        return data.to_numpy()

    def plot(self, sample: dict[str, Tensor], suptitle: str | None = None) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        image = sample["image"].take(self.rgb_indices, axis=0)
        image = np.transpose(image, (1, 2, 0))
        image = (image - image.min(axis=(0, 1))) * (1 / image.max(axis=(0, 1)))
        image = np.clip(image, 0, 1)

        label_mask = sample["mask"]
        label_mask = np.transpose(label_mask, (1, 2, 0))

        showing_predictions = "prediction" in sample
        if showing_predictions:
            prediction_mask = sample["prediction"]

        return self._plot_sample(
            image,
            label_mask,
            prediction=prediction_mask if showing_predictions else None,
            suptitle=suptitle,
        )

    @staticmethod
    def _plot_sample(image, label, num_classes, prediction=None, suptitle=None, class_names=None):
        num_images = 5 if prediction else 4
        fig, ax = plt.subplots(1, num_images, figsize=(8, 6))

        # for legend
        ax[0].axis("off")

        norm = mpl.colors.Normalize(vmin=0, vmax=num_classes - 1)
        ax[1].axis("off")
        ax[1].title.set_text("Image")
        ax[1].imshow(image)

        ax[2].axis("off")
        ax[2].title.set_text("Ground Truth Mask")
        ax[2].imshow(label, cmap="jet", norm=norm)

        ax[3].axis("off")
        ax[3].title.set_text("GT Mask on Image")
        ax[3].imshow(image)
        ax[3].imshow(label, cmap="jet", alpha=0.3, norm=norm)

        if prediction:
            ax[4].title.set_text("Predicted Mask")
            ax[4].imshow(prediction, cmap="jet", norm=norm)

        cmap = plt.get_cmap("jet")
        legend_data = []
        for i, _ in enumerate(range(num_classes)):
            class_name = class_names[i] if class_names else str(i)
            data = [i, cmap(norm(i)), class_name]
            legend_data.append(data)
        handles = [Rectangle((0, 0), 1, 1, color=tuple(v for v in c)) for k, c, n in legend_data]
        labels = [n for k, c, n in legend_data]
        ax[0].legend(handles, labels, loc="center")
        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig


class FireScarsHLS(RasterDataset):
    """RasterDataset implementation for fire scars input images."""

    filename_glob = "subsetted*_merged.tif"
    filename_regex = r"subsetted_512x512_HLS\..30\..{6}\.(?P<date>[0-9]*)\.v1.4_merged.tif"
    date_format = "%Y%j"
    is_image = True
    separate_files = False
    all_bands = dataclasses.field(default_factory=["B02", "B03", "B04", "B8A", "B11", "B12"])
    rgb_bands = dataclasses.field(default_factory=["B04", "B03", "B02"])


class FireScarsSegmentationMask(RasterDataset):
    """RasterDataset implementation for fire scars segmentation mask.
    Can be easily merged with input images using the & operator.
    """

    filename_glob = "subsetted*.mask.tif"
    filename_regex = r"subsetted_512x512_HLS\..30\..{6}\.(?P<date>[0-9]*)\.v1.4.mask.tif"
    date_format = "%Y%j"
    is_image = False
    separate_files = False



from collections.abc import Callable

import torch
from segmentation_models_pytorch.base import SegmentationModel
from torch import nn

from terratorch.models.heads import ClassificationHead
from terratorch.models.model import Model, ModelOutput


def freeze_module(module: nn.Module):
    for param in module.parameters():
        param.requires_grad_(False)


class ScalarOutputModel(Model, SegmentationModel):
    """Model that encapsulates encoder and decoder and heads for a scalar output
    Expects decoder to have a "forward_features" method, an embed_dims property
    and optionally a "prepare_features_for_image_model" method.
    """

    def __init__(
        self,
        task: str,
        encoder: nn.Module,
        decoder: nn.Module,
        head_kwargs: dict,
        auxiliary_heads: dict[str, nn.Module] | None = None,
        prepare_features_for_image_model: Callable | None = None,
    ) -> None:
        """Constructor

        Args:
            task (str): Task to be performed. Must be "classification".
            encoder (nn.Module): Encoder to be used
            decoder (nn.Module): Decoder to be used
            head_kwargs (dict): Arguments to be passed at instantiation of the head.
            auxiliary_heads (dict[str, nn.Module] | None, optional): Names mapped to auxiliary heads. Defaults to None.
            prepare_features_for_image_model (Callable | None, optional): Function applied to encoder outputs.
                Defaults to None.
        """
        super().__init__()
        self.task = task
        self.encoder = encoder
        self.decoder = decoder
        self.head = self._get_head(task, decoder.output_embed_dim, head_kwargs)

        if auxiliary_heads is not None:
            aux_heads = {}
            for aux_head_to_be_instantiated in auxiliary_heads:
                aux_head: nn.Module = self._get_head(
                    task, aux_head_to_be_instantiated.decoder.output_embed_dim, **head_kwargs
                )
                aux_head = nn.Sequential(aux_head_to_be_instantiated.decoder, aux_head)
                aux_heads[aux_head_to_be_instantiated.name] = aux_head
        else:
            aux_heads = {}
        self.aux_heads = nn.ModuleDict(aux_heads)

        self.prepare_features_for_image_model = prepare_features_for_image_model

    def freeze_encoder(self):
        freeze_module(self.encoder)

    def freeze_decoder(self):
        freeze_module(self.encoder)
        freeze_module(self.head)

    # TODO: do this properly
    def check_input_shape(self, x: torch.Tensor) -> bool:  # noqa: ARG002
        return True

    def forward(self, x: torch.Tensor) -> ModelOutput:
        """Sequentially pass `x` through model`s encoder, decoder and heads"""

        self.check_input_shape(x)
        features = self.encoder(x)

        # some models need their features reshaped

        if self.prepare_features_for_image_model:
            prepare = self.prepare_features_for_image_model
        else:
            prepare = getattr(self.encoder, "prepare_features_for_image_model", lambda x: x)
        features = prepare(features)

        decoder_output = self.decoder([f.clone() for f in features])
        mask = self.head(decoder_output)
        aux_outputs = {}
        for name, decoder in self.aux_heads.items():
            aux_output = decoder([f.clone() for f in features])
            aux_outputs[name] = aux_output
        return ModelOutput(output=mask, auxiliary_heads=aux_outputs)

    def _get_head(self, task: str, input_embed_dim: int, head_kwargs: dict):
        if task == "classification":
            if "num_classes" not in head_kwargs:
                msg = "num_classes must be defined for classification task"
                raise Exception(msg)
            return ClassificationHead(input_embed_dim, **head_kwargs)
        msg = "Task must be classification."
        raise Exception(msg)



import numpy as np
from collections.abc import Sequence
import torch
from pathlib import Path
import json
import h5py

import albumentations as A
from albumentations.pytorch import ToTensorV2

from torchgeo.datasets import NonGeoDataset
from terratorch.datasets.utils import to_tensor

import matplotlib.pyplot as plt
from matplotlib import colormaps
import matplotlib.patches as mpatches
from matplotlib.colors import Normalize

class MNeonTreeNonGeo(NonGeoDataset):
    all_band_names = ("BLUE", "CANOPY_HEIGHT_MODEL", "GREEN", "NEON", "RED")

    rgb_bands = ("RED","GREEN","BLUE")

    def __init__(self, data_root: str, bands: Sequence[str] = rgb_bands, transform: A.Compose | None = None, split="train") -> None:
        super().__init__()
        if split not in ["train", "test", "val"]:
            msg = "Split must be one of train, test, val."
            raise Exception(msg)
        if split == "val":
            split = "valid"
        
        self.transform = transform if transform else lambda **batch: to_tensor(batch)
        self._validate_bands(bands)
        self.bands = bands
        self.band_indices = np.array(
            [self.all_band_names.index(b) for b in bands if b in self.all_band_names]
        )
        self.split = split
        data_root = Path(data_root)
        self.data_directory = data_root / "m-NeonTree"

        partition_file = self.data_directory / "default_partition.json"
        with open(partition_file, 'r') as file:
            partitions = json.load(file)

        if split not in partitions:
            raise ValueError(f"Split '{split}' not found.")

        self.image_files = [self.data_directory / (filename + ".hdf5") for filename in partitions[split]]


    def __getitem__(self, index: int) -> dict[str, torch.Tensor]:
        file_path = self.image_files[index]
        
        with h5py.File(file_path, 'r') as h5file:
            keys = sorted(h5file.keys())
            keys = np.array([key for key in keys if key != 'label'])[self.band_indices]
            bands = [np.array(h5file[key]) for key in keys]            
            
            image = np.stack(bands, axis=-1)
            mask = np.array(h5file['label'])
                    
        output =  {
            "image": image.astype(np.float32),
            "mask": mask
        }

        output = self.transform(**output)
        output['mask'] = output['mask'].long()

        return output
    
    def _validate_bands(self, bands: Sequence[str]) -> None:
        assert isinstance(bands, Sequence), "'bands' must be a sequence"
        for band in bands:
            if band not in self.all_band_names:
                raise ValueError(f"'{band}' is an invalid band name.")

    def __len__(self):
        return len(self.image_files)
        
    def plot(self, arg, suptitle: str | None = None) -> None:
        if isinstance(arg, int):
            sample = self.__getitem__(arg)
        elif isinstance(arg, dict):
            sample = arg
        else:
            raise TypeError("Argument must be an integer index or a sample dictionary.")
            
        showing_predictions = sample['prediction'] if "prediction" in sample else None
        
        return self.plot_sample(sample, showing_predictions, suptitle)

    def plot_sample(self, sample, prediction=None, suptitle: str | None = None, class_names=None):
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise ValueError("Dataset doesn't contain some of the RGB bands")
        
        image = sample["image"].numpy()  
        image = image[rgb_indices, :, :]
        image = np.transpose(image, (1, 2, 0))  
        image = (image - image.min(axis=(0, 1))) * (1 / image.max(axis=(0, 1)))
        image = np.clip(image, 0, 1)
        
        mask = sample["mask"].numpy()
        num_classes = len(np.unique(mask))  
        
        num_images = 4 if prediction is not None else 3
        fig, ax = plt.subplots(1, num_images, figsize=(num_images * 4, 4), tight_layout=True)
        
        cmap = colormaps["jet"]
        norm = Normalize(vmin=0, vmax=num_classes - 1)
        
        ax[0].imshow(image)  
        ax[0].set_title("Image")
        ax[0].axis("off")
        
        ax[1].imshow(mask, cmap=cmap, norm=norm)
        ax[1].set_title("Ground Truth Mask")
        ax[1].axis("off")
        
        ax[2].imshow(image)  
        ax[2].imshow(mask, cmap=cmap, alpha=0.3, norm=norm) 
        ax[2].set_title("GT Mask on Image")
        ax[2].axis("off")
        
        if prediction is not None:
            prediction = prediction.numpy()
            ax[3].imshow(prediction, cmap=cmap, norm=norm)
            ax[3].set_title("Predicted Mask")
            ax[3].axis("off")
        
        if class_names:
            legend_handles = [mpatches.Patch(color=cmap(i), label=class_names[i]) for i in range(num_classes)]
            ax[0].legend(handles=legend_handles, bbox_to_anchor=(1.05, 1), loc='upper left')
        
        if suptitle:
            plt.suptitle(suptitle)
        
        plt.show()

        return fig
        


from typing import Any, Iterable
import torch

import albumentations as A
import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential
from terratorch.datasets import MEuroSATNonGeo
from terratorch.datamodules.utils import wrap_in_compose_is_list

MEANS = {
    "COASTAL_AEROSOL": 1355.5426, 
    "BLUE": 1113.8855, 
    "GREEN": 1035.7394,  
    "RED": 928.2619, 
    "RED_EDGE_1": 1188.2629, 
    "RED_EDGE_2": 2032.7325,
    "RED_EDGE_3": 2416.5286, 
    "NIR_BROAD": 2342.5396,  
    "NIR_NARROW": 748.9036,   
    "WATER_VAPOR": 12.0419, 
    "CIRRUS": 1810.1284, 
    "SWIR_1": 1101.3801,
    "SWIR_2": 2644.5996
}

STDS = {
    "COASTAL_AEROSOL": 68.9288, 
    "BLUE": 160.0012, 
    "GREEN": 194.6687, 
    "RED": 286.8012, 
    "RED_EDGE_1": 236.6991, 
    "RED_EDGE_2": 372.3853, 
    "RED_EDGE_3": 478.1329,
    "NIR_BROAD": 556.7527, 
    "NIR_NARROW": 102.5583,   
    "WATER_VAPOR": 1.2167, 
    "CIRRUS": 392.9388, 
    "SWIR_1": 313.7339, 
    "SWIR_2": 526.7788
}

class MEuroSATNonGeoDataModule(NonGeoDataModule):
    def __init__(
        self, 
        batch_size: int = 8, 
        num_workers: int = 0, 
        data_root: str = "./",
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        **kwargs: Any
    ) -> None:

        super().__init__(MEuroSATNonGeo, batch_size, num_workers, **kwargs)
        
        bands = kwargs.get("bands", MEuroSATNonGeo.all_band_names)
        self.means = torch.tensor([MEANS[b] for b in bands])
        self.stds = torch.tensor([STDS[b] for b in bands])
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.data_root = data_root
        self.aug = AugmentationSequential(K.Normalize(self.means, self.stds), data_keys=["image"]) if aug is None else aug
    
    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(  
                split="train", data_root=self.data_root, transform=self.train_transform, **self.kwargs
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="val", data_root=self.data_root, transform=self.val_transform, **self.kwargs
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",data_root=self.data_root, transform=self.test_transform, **self.kwargs
            )



import os

import pytest
import torch

from terratorch.models import PrithviModelFactory
from terratorch.models.backbones.prithvi_vit import PRETRAINED_BANDS

#from terratorch.models.backbones.prithvi_vit import default_cfgs as vit_default_cfgs
from terratorch.models.model import AuxiliaryHead

NUM_CHANNELS = 6
NUM_CLASSES = 2
EXPECTED_SEGMENTATION_OUTPUT_SHAPE = (1, NUM_CLASSES, 224, 224)
EXPECTED_REGRESSION_OUTPUT_SHAPE = (1, 224, 224)
EXPECTED_CLASSIFICATION_OUTPUT_SHAPE = (1, NUM_CLASSES)


@pytest.fixture(scope="session")
def model_factory() -> PrithviModelFactory:
    return PrithviModelFactory()


@pytest.fixture(scope="session")
def model_input() -> torch.Tensor:
    return torch.ones((1, NUM_CHANNELS, 224, 224))

@pytest.mark.parametrize("backbone", ["prithvi_vit_100", "prithvi_vit_300"])
def test_create_classification_model(backbone, model_factory: PrithviModelFactory, model_input):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        in_channels=NUM_CHANNELS,
        bands=PRETRAINED_BANDS,
        pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

@pytest.mark.parametrize("backbone", ["prithvi_vit_100", "prithvi_vit_300"])
def test_create_classification_model_no_in_channels(backbone, model_factory: PrithviModelFactory, model_input):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        bands=PRETRAINED_BANDS,
        pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

@pytest.mark.parametrize("backbone", ["prithvi_vit_100"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_segmentation_model(backbone, decoder, model_factory: PrithviModelFactory, model_input):
    model = model_factory.build_model(
        "segmentation",
        backbone=backbone,
        decoder=decoder,
        in_channels=NUM_CHANNELS,
        bands=PRETRAINED_BANDS,
        pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_SEGMENTATION_OUTPUT_SHAPE

@pytest.mark.parametrize("backbone", ["prithvi_vit_100"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_segmentation_model_no_in_channels(backbone, decoder, model_factory: PrithviModelFactory, model_input):
    model = model_factory.build_model(
        "segmentation",
        backbone=backbone,
        decoder=decoder,
        bands=PRETRAINED_BANDS,
        pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_SEGMENTATION_OUTPUT_SHAPE


@pytest.mark.parametrize("backbone", ["prithvi_vit_100"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_segmentation_model_with_aux_heads(backbone, decoder, model_factory: PrithviModelFactory, model_input):
    aux_heads_name = ["first_aux", "second_aux"]
    model = model_factory.build_model(
        "segmentation",
        backbone=backbone,
        decoder=decoder,
        in_channels=NUM_CHANNELS,
        bands=PRETRAINED_BANDS,
        pretrained=False,
        num_classes=NUM_CLASSES,
        aux_decoders=[AuxiliaryHead(name, "FCNDecoder", None) for name in aux_heads_name],
    )
    model.eval()

    with torch.no_grad():
        model_output = model(model_input)
        assert model_output.output.shape == EXPECTED_SEGMENTATION_OUTPUT_SHAPE

        assert len(model_output.auxiliary_heads.keys() & aux_heads_name) == len(aux_heads_name)
        for _, output in model_output.auxiliary_heads.items():
            assert output.shape == EXPECTED_SEGMENTATION_OUTPUT_SHAPE


@pytest.mark.parametrize("backbone", ["prithvi_vit_100"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_regression_model(backbone, decoder, model_factory: PrithviModelFactory, model_input):
    model = model_factory.build_model(
        "regression",
        backbone=backbone,
        decoder=decoder,
        in_channels=NUM_CHANNELS,
        bands=PRETRAINED_BANDS,
        pretrained=False,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_REGRESSION_OUTPUT_SHAPE

@pytest.mark.parametrize("backbone", ["prithvi_vit_100"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_regression_model_no_in_channels(backbone, decoder, model_factory: PrithviModelFactory, model_input):
    model = model_factory.build_model(
        "regression",
        backbone=backbone,
        decoder=decoder,
        bands=PRETRAINED_BANDS,
        pretrained=False,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_REGRESSION_OUTPUT_SHAPE

@pytest.mark.parametrize("backbone", ["prithvi_vit_100"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_regression_model_with_aux_heads(backbone, decoder, model_factory: PrithviModelFactory, model_input):
    aux_heads_name = ["first_aux", "second_aux"]
    model = model_factory.build_model(
        "regression",
        backbone=backbone,
        decoder=decoder,
        in_channels=NUM_CHANNELS,
        bands=PRETRAINED_BANDS,
        pretrained=False,
        aux_decoders=[AuxiliaryHead(name, "FCNDecoder", None) for name in aux_heads_name],
    )
    model.eval()

    with torch.no_grad():
        model_output = model(model_input)
        assert model_output.output.shape == EXPECTED_REGRESSION_OUTPUT_SHAPE

        assert len(model_output.auxiliary_heads.keys() & aux_heads_name) == len(aux_heads_name)
        for _, output in model_output.auxiliary_heads.items():
            assert output.shape == EXPECTED_REGRESSION_OUTPUT_SHAPE


@pytest.mark.parametrize("backbone", ["prithvi_vit_100"])
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_model_with_extra_bands(backbone, decoder, model_factory: PrithviModelFactory):
    model = model_factory.build_model(
        "segmentation",
        backbone=backbone,
        decoder=decoder,
        in_channels=NUM_CHANNELS + 1,
        bands=[*PRETRAINED_BANDS, 7],  # add an extra band
        pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()
    model_input = torch.ones((1, NUM_CHANNELS + 1, 224, 224))
    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_SEGMENTATION_OUTPUT_SHAPE



import os

import pytest
import torch
from _pytest.tmpdir import TempPathFactory

from terratorch.datasets import GenericNonGeoPixelwiseRegressionDataset, GenericNonGeoSegmentationDataset

REGRESSION_IMAGE_PATH = "tests/regression_test_input.tif"
REGRESSION_LABEL_PATH = "tests/regression_test_label.tif"
SEGMENTATION_IMAGE_PATH = "tests/segmentation_test_input.tif"
SEGMENTATION_LABEL_PATH = "tests/segmentation_test_label.tif"
NUM_CLASSES_SEGMENTATION = 2


@pytest.fixture(scope="session")
def split_file_path(tmp_path_factory):
    split_file_path = tmp_path_factory.mktemp("split") / "split.txt"
    with open(split_file_path, "w") as f:
        for i in range(5):
            f.write(f"{i}\n")
    return split_file_path


# test file discovery
class TestGenericRegressionDataset:
    @pytest.fixture(scope="class")
    def data_root_regression(self, tmp_path_factory: TempPathFactory):
        data_dir = tmp_path_factory.mktemp("data")
        image_dir_path = data_dir / "input_data"
        label_dir_path = data_dir / "label_data"
        os.mkdir(image_dir_path)
        os.mkdir(label_dir_path)
        for i in range(10):
            os.symlink(REGRESSION_IMAGE_PATH, image_dir_path / f"{i}_img.tif")
            os.symlink(REGRESSION_LABEL_PATH, label_dir_path / f"{i}_label.tif")

        # add a few with no suffix
        for i in range(10, 15):
            os.symlink(REGRESSION_IMAGE_PATH, image_dir_path / f"{i}.tif")
            os.symlink(REGRESSION_LABEL_PATH, label_dir_path / f"{i}.tif")
        return data_dir

    @pytest.fixture(scope="class")
    def regression_dataset(self, data_root_regression, split_file_path):
        return GenericNonGeoPixelwiseRegressionDataset(
            data_root_regression,
            image_grep="input_data/*_img.tif",
            label_grep="label_data/*_label.tif",
            split=split_file_path,
        )

    def test_file_discovery_generic_regression_dataset(self, regression_dataset):
        assert len(regression_dataset) == 5

    def test_data_type_regression_float_float(self, regression_dataset):
        assert torch.is_floating_point(regression_dataset[0]["image"])
        assert torch.is_floating_point(regression_dataset[0]["mask"])


class TestGenericSegmentationDataset:
    @pytest.fixture(scope="class")
    def data_root_segmentation(self, tmp_path_factory: TempPathFactory):
        data_dir = tmp_path_factory.mktemp("data")
        image_dir_path = data_dir / "input_data"
        label_dir_path = data_dir / "label_data"
        os.mkdir(image_dir_path)
        os.mkdir(label_dir_path)
        for i in range(10):
            os.symlink(SEGMENTATION_IMAGE_PATH, image_dir_path / f"{i}_img.tif")
            os.symlink(SEGMENTATION_LABEL_PATH, label_dir_path / f"{i}_label.tif")

        # add a few with no suffix
        for i in range(10, 15):
            os.symlink(SEGMENTATION_IMAGE_PATH, image_dir_path / f"{i}.tif")
            os.symlink(SEGMENTATION_LABEL_PATH, label_dir_path / f"{i}.tif")
        return data_dir

    @pytest.fixture(scope="class")
    def segmentation_dataset(self, data_root_segmentation, split_file_path):
        return GenericNonGeoSegmentationDataset(
            data_root_segmentation,
            image_grep="input_data/*_img.tif",
            label_grep="label_data/*_label.tif",
            num_classes=NUM_CLASSES_SEGMENTATION,
            split=split_file_path,
        )

    def test_file_discovery_generic_segmentation_dataset(self, segmentation_dataset):
        assert len(segmentation_dataset) == 5

    def test_data_type_regression_float_long(self, segmentation_dataset):
        assert torch.is_floating_point(segmentation_dataset[0]["image"])
        assert not torch.is_floating_point(segmentation_dataset[0]["mask"])



import numpy as np
from collections.abc import Sequence
import matplotlib.pyplot as plt
import torch
from pathlib import Path
import json
import h5py

import albumentations as A
from albumentations.pytorch import ToTensorV2

from torchgeo.datasets import NonGeoDataset
from terratorch.datasets.utils import to_tensor


class MBigEarthNonGeo(NonGeoDataset):

    all_band_names = (
        "COASTAL_AEROSOL",
        "BLUE",
        "GREEN",
        "RED",
        "RED_EDGE_1",
        "RED_EDGE_2",
        "RED_EDGE_3",
        "NIR_BROAD",
        "NIR_NARROW",
        "WATER_VAPOR",
        "SWIR_1",
        "SWIR_2",
        "CLOUD_PROBABILITY"
    )

    rgb_bands = ("RED", "GREEN", "BLUE")

    BAND_SETS = {"all": all_band_names, "rgb": rgb_bands}

    def __init__(self, data_root: str, bands: Sequence[str] = BAND_SETS["all"], transform: A.Compose | None = None, split="train", **kwargs: any) -> None:
        super().__init__()
        if split not in ["train", "test", "val"]:
            msg = "Split must be one of train, test, val."
            raise Exception(msg)
        if split == "val":
            split = "valid"
            
        self.transform = transform if transform else lambda **batch: to_tensor(batch)
        self._validate_bands(bands)
        self.bands = bands
        self.band_indices = np.array(
            [self.all_band_names.index(b) for b in bands if b in self.all_band_names]
        )
        self.split = split
        data_root = Path(data_root)
        self.data_directory = data_root / "m-bigearthnet"
        
        label_map_file = self.data_directory / "label_stats.json"
        with open(label_map_file, 'r') as file:
            self.label_map = json.load(file)

        partition_file = self.data_directory / "default_partition.json"
        with open(partition_file, 'r') as file:
            partitions = json.load(file)

        if split not in partitions:
            raise ValueError(f"Split '{split}' not found.")

        self.image_files = [self.data_directory / (filename + ".hdf5") for filename in partitions[split]]


    def __getitem__(self, index: int) -> dict[str, torch.Tensor]:
        file_path = self.image_files[index]
        image_id = file_path.stem 

        with h5py.File(file_path, 'r') as h5file:
            keys = sorted(h5file.keys())
            keys = np.array([key for key in keys if key != 'label'])[self.band_indices]
            bands = [np.array(h5file[key]) for key in keys]
            
            image = np.stack(bands, axis=-1)
        
        labels_vector = self.label_map[image_id]
        labels_tensor = torch.tensor(labels_vector, dtype=torch.float)

        output =  {
            "image": image,
            "label": labels_tensor
        }

        output = self.transform(**output)

        return output
    
    def _validate_bands(self, bands: Sequence[str]) -> None:
        assert isinstance(bands, Sequence), "'bands' must be a sequence"
        for band in bands:
            if band not in self.all_band_names:
                raise ValueError(f"'{band}' is an invalid band name.")
            
    def __len__(self):
        return len(self.image_files)
        
    def plot(self, arg, suptitle: str | None = None) -> None:
        if isinstance(arg, int):
            sample = self.__getitem__(arg)
        elif isinstance(arg, dict):
            sample = arg
        else:
            raise TypeError("Argument must be an integer index or a sample dictionary.")

        image = sample["image"].numpy()
        labels = sample["label"].numpy()

        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise ValueError("Dataset doesn't contain some of the RGB bands")

        rgb_image = image[rgb_indices, :, :]
        rgb_image = np.transpose(rgb_image, (1, 2, 0))
        rgb_image = (rgb_image - np.min(rgb_image)) / (np.max(rgb_image) - np.min(rgb_image))
    
        active_labels = [i for i, label in enumerate(labels) if label == 1]

        self._plot_sample(
            image=rgb_image,
            label_indices=active_labels,
            suptitle=suptitle
        )
            
    @staticmethod
    def _plot_sample(image, label_indices, suptitle=None) -> None:
        fig, ax = plt.subplots(figsize=(6, 6))
        ax.imshow(image)
        ax.axis('off')

        title = f'Active Labels: {label_indices}'
        if suptitle:
            title = f'{suptitle} - {title}'
        ax.set_title(title)
    
        return fig




from typing import Any, Iterable
import torch

import albumentations as A
import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential
from terratorch.datasets import MForestNetNonGeo
from terratorch.datamodules.utils import wrap_in_compose_is_list


MEANS = {
    "BLUE": 72.852258, 
    "GREEN": 83.677155, 
    "RED": 77.58181, 
    "NIR": 123.987442, 
    "SWIR_1": 91.536942,
    "SWIR_2": 74.719202
}

STDS = {
    "BLUE": 15.837172547567825,
    "GREEN": 14.788812599596188,
    "RED": 16.100543441881086,
    "NIR": 16.35234883118129,
    "SWIR_1": 13.7882739778638,
    "SWIR_2": 12.69131413539181
}

class MForestNetNonGeoDataModule(NonGeoDataModule):
    def __init__(
        self, 
        batch_size: int = 8, 
        num_workers: int = 0, 
        data_root: str = "./",
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        **kwargs: Any
    ) -> None:

        super().__init__(MForestNetNonGeo, batch_size, num_workers, **kwargs)
        
        bands = kwargs.get("bands", MForestNetNonGeo.all_band_names)
        self.means = torch.tensor([MEANS[b] for b in bands])
        self.stds = torch.tensor([STDS[b] for b in bands])
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.data_root = data_root
        self.aug = AugmentationSequential(K.Normalize(self.means, self.stds), data_keys=["image"]) if aug is None else aug
    
    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(  
                split="train", data_root=self.data_root, transform=self.train_transform, **self.kwargs
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="val", data_root=self.data_root, transform=self.val_transform, **self.kwargs
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",data_root=self.data_root, transform=self.test_transform, **self.kwargs
            )



from abc import ABC, abstractmethod
from dataclasses import dataclass

from torch import Tensor, nn

FACTORY_REGISTRY = {}


@dataclass
class ModelOutput:
    output: Tensor
    auxiliary_heads: dict[str, Tensor] = None


class Model(ABC):
    @abstractmethod
    def freeze_encoder(self):
        pass

    @abstractmethod
    def freeze_decoder(self):
        pass

    @abstractmethod
    def forward(self) -> ModelOutput:
        pass


class ModelFactory(ABC):
    @abstractmethod
    def build_model() -> Model:
        pass


def get_factory(factory_name: str) -> ModelFactory:
    if factory_name not in FACTORY_REGISTRY:
        msg = f"Factory with name {factory_name} does not exist. Choose one of {list(FACTORY_REGISTRY.keys())}"
        raise Exception(msg)
    return FACTORY_REGISTRY[factory_name]()


def register_factory(factory_class: type) -> None:
    FACTORY_REGISTRY[factory_class.__name__] = factory_class
    return factory_class


@dataclass
class AuxiliaryHead:
    """Class containing all information to create auxiliary heads.

    Args:
        name (str): Name of the head. Should match the name given to the auxiliary loss.
        decoder (str): Name of the decoder class to be used.
        decoder_args (dict | None): parameters to be passed to the decoder constructor.
            Parameters for the decoder should be prefixed with `decoder_`.
            Parameters for the head should be prefixed with `head_`.
    """

    name: str
    decoder: str
    decoder_args: dict | None


@dataclass
class AuxiliaryHeadWithDecoderWithoutInstantiatedHead:
    """Intermediary class containing the instantiated decoder without the instantiated head.

    Args:
        name (str): Name of the head. Should match the name given to the auxiliary loss.
        decoder (nn.Module): Instantiated decoder.
        head_args (dict | None): parameters to be passed to the head constructor.
    """

    name: str
    decoder: nn.Module
    head_args: dict | None



from pathlib import Path
from typing import Any

import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential

from terratorch.datasets import Sen1Floods11NonGeo

MEANS = [
    0.033349706741586264,
    0.05701185520536176,
    0.05889748132001316,
    0.2323245113436119,
    0.1972854853760658,
    0.11944914225186566,
]

STDS = [
    0.02269135568823774,
    0.026807560223070237,
    0.04004109844362779,
    0.07791732423672691,
    0.08708738838140137,
    0.07241979477437814,
]


class Sen1Floods11NonGeoDataModule(NonGeoDataModule):
    """NonGeo Fire Scars data module implementation"""

    def __init__(self, batch_size: int = 4, num_workers: int = 0, **kwargs: Any) -> None:
        super().__init__(Sen1Floods11NonGeo, batch_size, num_workers, **kwargs)
        self.aug = AugmentationSequential(K.Normalize(MEANS, STDS), data_keys=["image", "mask"])



from typing import Any
import torch

import albumentations as A
import kornia.augmentation as K  # noqa: N812
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential
from terratorch.datasets import MBrickKilnNonGeo
from terratorch.datamodules.utils import wrap_in_compose_is_list


MEANS = {
    "COASTAL_AEROSOL": 574.7587880700896,
    "BLUE": 674.3473615470523,
    "GREEN": 886.3656479311578,
    "RED": 815.0945462528913,
    "RED_EDGE_1": 1128.8088426870465,
    "RED_EDGE_2": 1934.450471876027,
    "RED_EDGE_3": 2045.7652282437202,
    "NIR_BROAD": 2012.744587807115,
    "NIR_NARROW": 1608.6255233989034,
    "WATER_VAPOR": 1129.8171906000355,
    "CIRRUS": 83.27188605598549,
    "SWIR_1": 90.54924599052214,
    "SWIR_2": 68.98768652434848
}

STDS = {
    "COASTAL_AEROSOL": 193.60631504991184,
    "BLUE": 238.75447480113132,
    "GREEN": 276.9631260242207,
    "RED": 361.15060137326634,
    "RED_EDGE_1": 364.5888078793488,
    "RED_EDGE_2": 724.2707123576525,
    "RED_EDGE_3": 819.653063972575,
    "NIR_BROAD": 794.3652427593881,
    "NIR_NARROW": 800.8538290702304,
    "WATER_VAPOR": 704.0219637458916,
    "CIRRUS": 36.355745901131705,
    "SWIR_1": 28.004671947623894,
    "SWIR_2": 24.268892726362033
}

class MBrickKilnNonGeoDataModule(NonGeoDataModule):

    def __init__(
        self, 
        batch_size: int = 8, 
        num_workers: int = 0, 
        data_root: str = "./",
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        **kwargs: Any
    ) -> None:

        super().__init__(MBrickKilnNonGeo, batch_size, num_workers, **kwargs)
        
        bands = kwargs.get("bands", MBrickKilnNonGeo.all_band_names)
        self.means = torch.tensor([MEANS[b] for b in bands])
        self.stds = torch.tensor([STDS[b] for b in bands])
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.data_root = data_root
        self.aug = AugmentationSequential(K.Normalize(self.means, self.stds), data_keys=["image"]) if aug is None else aug
    
    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(  
                split="train", data_root=self.data_root, transform=self.train_transform, **self.kwargs
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="val", data_root=self.data_root, transform=self.val_transform, **self.kwargs
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",data_root=self.data_root, transform=self.test_transform, **self.kwargs
            )



import torch
import torch.nn.functional as F  # noqa: N812
from torch import Tensor, nn

"""
Adapted from https://github.com/yassouali/pytorch-segmentation/blob/master/models/upernet.py
"""


class ConvModule(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding=0, inplace=False) -> None:  # noqa: FBT002
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, padding=padding, bias=False)
        self.norm = nn.BatchNorm2d(out_channels)
        self.act = nn.ReLU(inplace=inplace)

    def forward(self, x):
        return self.act(self.norm(self.conv(x)))


# class PSPModule(nn.Module):
#     # In the original inmplementation they use precise RoI pooling
#     # Instead of using adaptative average pooling
#     def __init__(self, in_channels: int, bin_sizes: list[int] | None = None):
#         super().__init__()
#         if bin_sizes is None:
#             bin_sizes = [1, 2, 3, 6]
#         out_channels = in_channels // len(bin_sizes)
#         self.stages = nn.ModuleList([self._make_stages(in_channels, out_channels, b_s) for b_s in bin_sizes])
#         self.bottleneck = nn.Sequential(
#             nn.Conv2d(
#                 in_channels + (out_channels * len(bin_sizes)),
#                 in_channels,
#                 kernel_size=3,
#                 padding=1,
#                 bias=False,
#             ),
#             nn.BatchNorm2d(in_channels),
#             nn.ReLU(inplace=True),
#             nn.Dropout2d(0.1),
#         )

#     def _make_stages(self, in_channels, out_channels, bin_sz):
#         prior = nn.AdaptiveAvgPool2d(output_size=bin_sz)
#         conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
#         bn = nn.BatchNorm2d(out_channels)
#         relu = nn.ReLU(inplace=True)
#         return nn.Sequential(prior, conv, bn, relu)

#     def forward(self, features):
#         h, w = features.size()[2], features.size()[3]
#         pyramids = [features]
#         pyramids.extend(
#             [F.interpolate(stage(features), size=(h, w), mode="bilinear", align_corners=True) for stage in self.stages]
#         )
#         output = self.bottleneck(torch.cat(pyramids, dim=1))
#         return output


# def up_and_add(x, y):
#     return F.interpolate(x, size=(y.size(2), y.size(3)), mode="bilinear", align_corners=True) + y


# class FPNFuse(nn.Module):
#     def __init__(self, feature_channels=None, fpn_out=256):
#         super().__init__()
#         if feature_channels is None:
#             feature_channels = [256, 512, 1024, 2048]
#         if not feature_channels[0] == fpn_out:
#             msg = f"First index of feature channel ({feature_channels[0]}) did not match fpn_out ({fpn_out})"
#             raise Exception(msg)
#         self.conv1x1 = nn.ModuleList([nn.Conv2d(ft_size, fpn_out, kernel_size=1) for ft_size in feature_channels[1:]])
#         self.smooth_conv = nn.ModuleList(
#             [nn.Conv2d(fpn_out, fpn_out, kernel_size=3, padding=1)] * (len(feature_channels) - 1)
#         )
#         self.conv_fusion = nn.Sequential(
#             nn.Conv2d(
#                 len(feature_channels) * fpn_out,
#                 fpn_out,
#                 kernel_size=3,
#                 padding=1,
#                 bias=False,
#             ),
#             nn.BatchNorm2d(fpn_out),
#             nn.ReLU(inplace=True),
#         )

#     def forward(self, features):
#         features[1:] = [conv1x1(feature) for feature, conv1x1 in zip(features[1:], self.conv1x1, strict=False)]
#         p = [up_and_add(features[i], features[i - 1]) for i in reversed(range(1, len(features)))]
#         p = [smooth_conv(x) for smooth_conv, x in zip(self.smooth_conv, p, strict=False)]
#         p = list(reversed(p))
#         p.append(features[-1])  # P = [P1, P2, P3, P4]
#         h, w = p[0].size(2), p[0].size(3)
#         p[1:] = [F.interpolate(feature, size=(h, w), mode="bilinear", align_corners=True) for feature in p[1:]]

#         x = self.conv_fusion(torch.cat(p, dim=1))
#         return x


# class UperNetDecoder(nn.Module):
#     def __init__(self, embed_dim: list[int]) -> None:
#         super().__init__()
#         self.embed_dim = embed_dim
#         self.output_embed_dim = embed_dim[0]
#         self.PPN = PSPModule(embed_dim[-1])
#         self.FPN = FPNFuse(embed_dim, fpn_out=self.output_embed_dim)

#     def forward(self, x: Tensor):
#         x = [f.clone() for f in x]
#         x[-1] = self.PPN(x[-1])
#         x = self.FPN(x)

#         return x


# Adapted from MMSegmentation
class UperNetDecoder(nn.Module):
    """UperNetDecoder. Adapted from MMSegmentation."""

    def __init__(
        self,
        embed_dim: list[int],
        pool_scales: tuple[int] = (1, 2, 3, 6),
        channels: int = 256,
        align_corners: bool = True,  # noqa: FBT001, FBT002
    ):
        """Constructor

        Args:
            embed_dim (list[int]): Input embedding dimension for each input.
            pool_scales (tuple[int], optional): Pooling scales used in Pooling Pyramid
                Module applied on the last feature. Default: (1, 2, 3, 6).
            channels (int, optional): Channels used in the decoder. Defaults to 256.
            align_corners (bool, optional): Whter to align corners in rescaling. Defaults to True.
        """
        super().__init__()
        self.embed_dim = embed_dim
        self.output_embed_dim = channels
        self.channels = channels
        self.align_corners = align_corners
        # PSP Module
        self.psp_modules = PPM(
            pool_scales,
            self.embed_dim[-1],
            self.channels,
            align_corners=self.align_corners,
        )
        self.bottleneck = ConvModule(
            self.embed_dim[-1] + len(pool_scales) * self.channels, self.channels, 3, padding=1, inplace=True
        )
        # FPN Module
        self.lateral_convs = nn.ModuleList()
        self.fpn_convs = nn.ModuleList()
        for embed_dim in self.embed_dim[:-1]:  # skip the top layer
            l_conv = ConvModule(
                embed_dim,
                self.channels,
                1,
                inplace=False,
            )
            fpn_conv = ConvModule(
                self.channels,
                self.channels,
                3,
                padding=1,
                inplace=False,
            )
            self.lateral_convs.append(l_conv)
            self.fpn_convs.append(fpn_conv)

        self.fpn_bottleneck = ConvModule(len(self.embed_dim) * self.channels, self.channels, 3, padding=1, inplace=True)

    def psp_forward(self, inputs):
        """Forward function of PSP module."""
        x = inputs[-1]
        psp_outs = [x]
        psp_outs.extend(self.psp_modules(x))
        psp_outs = torch.cat(psp_outs, dim=1)
        output = self.bottleneck(psp_outs)

        return output

    def forward(self, inputs):
        """Forward function for feature maps before classifying each pixel with
        Args:
            inputs (list[Tensor]): List of multi-level img features.

        Returns:
            feats (Tensor): A tensor of shape (batch_size, self.channels,
                H, W) which is feature map for last layer of decoder head.
        """
        # build laterals
        laterals = [lateral_conv(inputs[i]) for i, lateral_conv in enumerate(self.lateral_convs)]
        laterals.append(self.psp_forward(inputs))

        # build top-down path
        used_backbone_levels = len(laterals)
        for i in range(used_backbone_levels - 1, 0, -1):
            prev_shape = laterals[i - 1].shape[2:]
            laterals[i - 1] = laterals[i - 1] + torch.nn.functional.interpolate(
                laterals[i], size=prev_shape, mode="bilinear", align_corners=self.align_corners
            )

        # build outputs
        fpn_outs = [self.fpn_convs[i](laterals[i]) for i in range(used_backbone_levels - 1)]
        # append psp feature
        fpn_outs.append(laterals[-1])

        for i in range(used_backbone_levels - 1, 0, -1):
            fpn_outs[i] = torch.nn.functional.interpolate(
                fpn_outs[i], size=fpn_outs[0].shape[2:], mode="bilinear", align_corners=self.align_corners
            )
        fpn_outs = torch.cat(fpn_outs, dim=1)
        feats = self.fpn_bottleneck(fpn_outs)
        return feats


class PPM(nn.ModuleList):
    """Pooling Pyramid Module used in PSPNet."""

    def __init__(self, pool_scales, in_channels, channels, align_corners):
        """Constructor

        Args:
            pool_scales (tuple[int]): Pooling scales used in Pooling Pyramid
                Module.
            in_channels (int): Input channels.
            channels (int): Channels after modules, before conv_seg.
            align_corners (bool): align_corners argument of F.interpolate.
        """
        super().__init__()
        self.pool_scales = pool_scales
        self.align_corners = align_corners
        self.in_channels = in_channels
        self.channels = channels

        for pool_scale in pool_scales:
            self.append(
                nn.Sequential(
                    nn.AdaptiveAvgPool2d(pool_scale),
                    ConvModule(self.in_channels, self.channels, 1, inplace=True),
                )
            )

    def forward(self, x):
        """Forward function."""
        ppm_outs = []
        for ppm in self:
            ppm_out = ppm(x)
            upsampled_ppm_out = torch.nn.functional.interpolate(
                ppm_out, size=x.size()[2:], mode="bilinear", align_corners=self.align_corners
            )
            ppm_outs.append(upsampled_ppm_out)
        return ppm_outs



import albumentations as A
from typing import Iterable


def wrap_in_compose_is_list(transform_list):
    # set check shapes to false because of the multitemporal case
    return A.Compose(transform_list, is_check_shapes=False) if isinstance(transform_list, Iterable) else transform_list


from terratorch.datamodules.fire_scars import FireScarsNonGeoDataModule
from terratorch.datamodules.generic_pixel_wise_data_module import (
    GenericNonGeoPixelwiseRegressionDataModule,
    GenericNonGeoSegmentationDataModule,
)
from terratorch.datamodules.generic_scalar_label_data_module import (
    GenericNonGeoClassificationDataModule,
)

# GenericNonGeoRegressionDataModule,
from terratorch.datamodules.sen1floods11 import Sen1Floods11NonGeoDataModule
from terratorch.datamodules.torchgeo_data_module import TorchGeoDataModule, TorchNonGeoDataModule

# geobench classification datamodules
from terratorch.datamodules.m_eurosat import MEuroSATNonGeoDataModule
from terratorch.datamodules.m_bigearthnet import MBigEarthNonGeoDataModule
from terratorch.datamodules.m_brick_kiln import MBrickKilnNonGeoDataModule
from terratorch.datamodules.m_forestnet import MForestNetNonGeoDataModule
from terratorch.datamodules.m_so2sat import MSo2SatNonGeoDataModule
from terratorch.datamodules.m_pv4ger import MPv4gerNonGeoDataModule

# geobench segmentation datamodules
from terratorch.datamodules.m_cashew_plantation import MBeninSmallHolderCashewsNonGeoDataModule
from terratorch.datamodules.m_nz_cattle import MNzCattleNonGeoDataModule
from terratorch.datamodules.m_chesapeake_landcover import MChesapeakeLandcoverNonGeoDataModule
from terratorch.datamodules.m_pv4ger_seg import MPv4gerSegNonGeoDataModule
from terratorch.datamodules.m_SA_crop_type import MSACropTypeNonGeoDataModule
from terratorch.datamodules.m_neontree import MNeonTreeNonGeoDataModule

__all__ = (
    "GenericNonGeoSegmentationDataModule",
    "GenericNonGeoPixelwiseRegressionDataModule",
    "GenericNonGeoSegmentationDataModule",
    "GenericNonGeoClassificationDataModule",
    # "GenericNonGeoRegressionDataModule",
    "Sen1Floods11NonGeoDataModule",
    "FireScarsNonGeoDataModule",
    "TorchGeoDataModule",
    "TorchNonGeoDataModule",
    "MEuroSATNonGeoDataModule",
    "MBigEarthNonGeoDataModule",
    "MBrickKilnNonGeoDataModule",
    "MForestNetNonGeoDataModule",
    "MSo2SatNonGeoDataModule",
    "MPv4gerNonGeoDataModule",
    "MBeninSmallHolderCashewsNonGeoDataModule",
    "MNzCattleNonGeoDataModule",
    "MChesapeakeLandcoverNonGeoDataModule",
    "MPv4gerSegNonGeoDataModule",
    "MSACropTypeNonGeoDataModule",
    "MNeonTreeNonGeoDataModule"
)



"""
This is just an example of a possible structure to include SMP models
Right now it always returns a UNET, but could easily be extended to many of the models provided by SMP.
"""

import segmentation_models_pytorch as smp
from torch import nn

from terratorch.models.model import Model, ModelFactory, ModelOutput, register_factory


@register_factory
class SMPModelFactory(ModelFactory):
    def build_model(
        self,
        task: str,
        backbone: str,
        decoder: str,
        in_channels: int,
        pretrained: str | bool | None = True,
        num_classes: int = 1,
        regression_relu: bool = False,
        **kwargs,
    ) -> Model:
        """Factory to create model based on SMP.

        Args:
            task (str): Must be "segmentation".
            backbone (str): Name of backbone.
            decoder (str): Decoder architecture. Currently only supports "unet".
            in_channels (int): Number of input channels.
            pretrained(str | bool): Which weights to use for the backbone. If true, will use "imagenet". If false or None, random weights. Defaults to True.
            num_classes (int): Number of classes.
            regression_relu (bool). Whether to apply a ReLU if task is regression. Defaults to False.

        Returns:
            Model: SMP model wrapped in SMPModelWrapper.
        """
        if task not in ["segmentation", "regression"]:
            msg = f"SMP models can only perform pixel wise tasks, but got task {task}"
            raise Exception(msg)
        # backbone_kwargs = _extract_prefix_keys(kwargs, "backbone_")
        if isinstance(pretrained, bool):
            if pretrained:
                pretrained = "imagenet"
            else:
                pretrained = None
        if decoder == "unet":
            model = smp.Unet(
                encoder_name=backbone, encoder_weights=pretrained, in_channels=in_channels, classes=num_classes
            )
        else:
            msg = "Only unet decoder implemented"
            raise NotImplementedError(msg)
        return SMPModelWrapper(
            model, relu=task == "regression" and regression_relu, squeeze_single_class=task == "regression"
        )


class SMPModelWrapper(Model, nn.Module):
    def __init__(self, smp_model, relu=False, squeeze_single_class=False) -> None:
        super().__init__()
        self.smp_model = smp_model
        self.final_act = nn.ReLU() if relu else nn.Identity()
        self.squeeze_single_class = squeeze_single_class

    def forward(self, *args, **kwargs):
        smp_output = self.smp_model(*args, **kwargs)
        smp_output = self.final_act(smp_output)
        if smp_output.shape[1] == 1 and self.squeeze_single_class:
            smp_output = smp_output.squeeze(1)
        return ModelOutput(smp_output)

    def freeze_encoder(self):
        raise NotImplementedError()

    def freeze_decoder(self):
        raise NotImplementedError()


def _extract_prefix_keys(d: dict, prefix: str) -> dict:
    extracted_dict = {}
    keys_to_del = []
    for k, v in d.items():
        if k.startswith(prefix):
            extracted_dict[k.split(prefix)[1]] = v
            keys_to_del.append(k)

    for k in keys_to_del:
        del d[k]

    return extracted_dict



"""
This module contains generic data modules for instantiation at runtime.
"""

from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any

import albumentations as A
import kornia.augmentation as K
import torch
from torch import Tensor
from torch.utils.data import DataLoader
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential

from terratorch.datasets import GenericNonGeoPixelwiseRegressionDataset, GenericNonGeoSegmentationDataset, HLSBands


def wrap_in_compose_is_list(transform_list):
    # set check shapes to false because of the multitemporal case
    return A.Compose(transform_list, is_check_shapes=False) if isinstance(transform_list, Iterable) else transform_list


# def collate_fn_list_dicts(batch):
#     metadata = []
#     for i in range(len(batch)):
#         metadata.append(batch[i].pop("metadata"))
#     batch = default_collate(batch)
#     # print(len(batch["image"]), len(batch["mask"]))
#     batch["metadata"] = metadata
#     # print(len(batch["metadata"]))
#     if len(batch["image"]) != len(batch["metadata"]):
#         print(len(batch["image"]), len(batch["metadata"]))
#     return batch


class Normalize(Callable):
    def __init__(self, means, stds):
        super().__init__()
        self.means = means
        self.stds = stds

    def __call__(self, batch):
        # min_value = self.means - 2 * self.stds
        # max_value = self.means + 2 * self.stds
        # img = (batch["image"] - min_value) / (max_value - min_value)
        # img = torch.clip(img, 0, 1)
        # batch["image"] = img
        # return batch
        image = batch["image"]
        if len(image.shape) == 5:
            means = torch.tensor(self.means, device=image.device).view(1, -1, 1, 1, 1)
            stds = torch.tensor(self.stds, device=image.device).view(1, -1, 1, 1, 1)
        elif len(image.shape) == 4:
            means = torch.tensor(self.means, device=image.device).view(1, -1, 1, 1)
            stds = torch.tensor(self.stds, device=image.device).view(1, -1, 1, 1)
        else:
            msg = f"Expected batch to have 5 or 4 dimensions, but got {len(image.shape)}"
            raise Exception(msg)
        batch["image"] = (image - means) / stds
        return batch


class GenericNonGeoSegmentationDataModule(NonGeoDataModule):
    """
    This is a generic datamodule class for instantiating data modules at runtime.
    Composes several [GenericNonGeoSegmentationDatasets][terratorch.datasets.GenericNonGeoSegmentationDataset]
    """

    def __init__(
        self,
        batch_size: int,
        num_workers: int,
        train_data_root: Path,
        val_data_root: Path,
        test_data_root: Path,
        img_grep: str,
        label_grep: str,
        means: list[float],
        stds: list[float],
        num_classes: int,
        predict_data_root: Path | None = None,
        train_label_data_root: Path | None = None,
        val_label_data_root: Path | None = None,
        test_label_data_root: Path | None = None,
        train_split: Path | None = None,
        val_split: Path | None = None,
        test_split: Path | None = None,
        ignore_split_file_extensions: bool = True,
        allow_substring_split_file: bool = True,
        dataset_bands: list[HLSBands | int] | None = None,
        predict_dataset_bands: list[HLSBands | int] | None = None,
        output_bands: list[HLSBands | int] | None = None,
        constant_scale: float = 1,
        rgb_indices: list[int] | None = None,
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        expand_temporal_dimension: bool = False,
        reduce_zero_label: bool = False,
        no_data_replace: float | None = None,
        no_label_replace: int | None = None,
        drop_last: bool = True,
        **kwargs: Any,
    ) -> None:
        """Constructor

        Args:
            batch_size (int): _description_
            num_workers (int): _description_
            train_data_root (Path): _description_
            val_data_root (Path): _description_
            test_data_root (Path): _description_
            predict_data_root (Path): _description_
            img_grep (str): _description_
            label_grep (str): _description_
            means (list[float]): _description_
            stds (list[float]): _description_
            num_classes (int): _description_
            train_label_data_root (Path | None, optional): _description_. Defaults to None.
            val_label_data_root (Path | None, optional): _description_. Defaults to None.
            test_label_data_root (Path | None, optional): _description_. Defaults to None.
            train_split (Path | None, optional): _description_. Defaults to None.
            val_split (Path | None, optional): _description_. Defaults to None.
            test_split (Path | None, optional): _description_. Defaults to None.
            ignore_split_file_extensions (bool, optional): Whether to disregard extensions when using the split
                file to determine which files to include in the dataset.
                E.g. necessary for Eurosat, since the split files specify ".jpg" but files are
                actually ".jpg". Defaults to True.
            allow_substring_split_file (bool, optional): Whether the split files contain substrings
                that must be present in file names to be included (as in mmsegmentation), or exact
                matches (e.g. eurosat). Defaults to True.
            dataset_bands (list[HLSBands | int] | None, optional): _description_. Defaults to None.
            predict_dataset_bands (list[HLSBands | int] | None, optional): _description_. Defaults to None.
            output_bands (list[HLSBands | int] | None, optional): _description_. Defaults to None.
            constant_scale (float, optional): _description_. Defaults to 1.
            rgb_indices (list[int] | None, optional): _description_. Defaults to None.
            train_transform (Albumentations.Compose | None): Albumentations transform
                to be applied to the train dataset.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            val_transform (Albumentations.Compose | None): Albumentations transform
                to be applied to the train dataset.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            test_transform (Albumentations.Compose | None): Albumentations transform
                to be applied to the train dataset.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            no_data_replace (float | None): Replace nan values in input images with this value. If none, does no replacement. Defaults to None.
            no_label_replace (int | None): Replace nan values in label with this value. If none, does no replacement. Defaults to None.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Defaults to False.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
            drop_last (bool): Drop the last batch if it is not complete. Defaults to True.
        """
        super().__init__(GenericNonGeoSegmentationDataset, batch_size, num_workers, **kwargs)
        self.num_classes = num_classes
        self.img_grep = img_grep
        self.label_grep = label_grep
        self.train_root = train_data_root
        self.val_root = val_data_root
        self.test_root = test_data_root
        self.predict_root = predict_data_root
        self.train_split = train_split
        self.val_split = val_split
        self.test_split = test_split
        self.ignore_split_file_extensions = ignore_split_file_extensions
        self.allow_substring_split_file = allow_substring_split_file
        self.constant_scale = constant_scale
        self.no_data_replace = no_data_replace
        self.no_label_replace = no_label_replace
        self.drop_last = drop_last

        self.train_label_data_root = train_label_data_root
        self.val_label_data_root = val_label_data_root
        self.test_label_data_root = test_label_data_root

        self.dataset_bands = dataset_bands
        self.predict_dataset_bands = predict_dataset_bands if predict_dataset_bands else dataset_bands
        self.output_bands = output_bands
        self.rgb_indices = rgb_indices
        self.expand_temporal_dimension = expand_temporal_dimension
        self.reduce_zero_label = reduce_zero_label

        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)

        # self.aug = AugmentationSequential(
        #     K.Normalize(means, stds),
        #     data_keys=["image"],
        # )
        self.aug = Normalize(means, stds)

        # self.aug = Normalize(means, stds)
        # self.collate_fn = collate_fn_list_dicts

    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(
                self.train_root,
                self.num_classes,
                image_grep=self.img_grep,
                label_grep=self.label_grep,
                label_data_root=self.train_label_data_root,
                split=self.train_split,
                ignore_split_file_extensions=self.ignore_split_file_extensions,
                allow_substring_split_file=self.allow_substring_split_file,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                rgb_indices=self.rgb_indices,
                transform=self.train_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                self.val_root,
                self.num_classes,
                image_grep=self.img_grep,
                label_grep=self.label_grep,
                label_data_root=self.val_label_data_root,
                split=self.val_split,
                ignore_split_file_extensions=self.ignore_split_file_extensions,
                allow_substring_split_file=self.allow_substring_split_file,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                rgb_indices=self.rgb_indices,
                transform=self.val_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                self.test_root,
                self.num_classes,
                image_grep=self.img_grep,
                label_grep=self.label_grep,
                label_data_root=self.test_label_data_root,
                split=self.test_split,
                ignore_split_file_extensions=self.ignore_split_file_extensions,
                allow_substring_split_file=self.allow_substring_split_file,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                rgb_indices=self.rgb_indices,
                transform=self.test_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
            )
        if stage in ["predict"] and self.predict_root:
            self.predict_dataset = self.dataset_class(
                self.predict_root,
                self.num_classes,
                dataset_bands=self.predict_dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                rgb_indices=self.rgb_indices,
                transform=self.test_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
            )

    def _dataloader_factory(self, split: str) -> DataLoader[dict[str, Tensor]]:
        """Implement one or more PyTorch DataLoaders.

        Args:
            split: Either 'train', 'val', 'test', or 'predict'.

        Returns:
            A collection of data loaders specifying samples.

        Raises:
            MisconfigurationException: If :meth:`setup` does not define a
                dataset or sampler, or if the dataset or sampler has length 0.
        """
        dataset = self._valid_attribute(f"{split}_dataset", "dataset")
        batch_size = self._valid_attribute(f"{split}_batch_size", "batch_size")
        return DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=split == "train",
            num_workers=self.num_workers,
            collate_fn=self.collate_fn,
            drop_last=split == "train" and self.drop_last,
        )


class GenericNonGeoPixelwiseRegressionDataModule(NonGeoDataModule):
    """This is a generic datamodule class for instantiating data modules at runtime.
    Composes several
    [GenericNonGeoPixelwiseRegressionDataset][terratorch.datasets.GenericNonGeoPixelwiseRegressionDataset]
    """

    def __init__(
        self,
        batch_size: int,
        num_workers: int,
        train_data_root: Path,
        val_data_root: Path,
        test_data_root: Path,
        means: list[float],
        stds: list[float],
        predict_data_root: Path | None = None,
        img_grep: str | None = "*",
        label_grep: str | None = "*",
        train_label_data_root: Path | None = None,
        val_label_data_root: Path | None = None,
        test_label_data_root: Path | None = None,
        train_split: Path | None = None,
        val_split: Path | None = None,
        test_split: Path | None = None,
        ignore_split_file_extensions: bool = True,
        allow_substring_split_file: bool = True,
        dataset_bands: list[HLSBands | int] | None = None,
        predict_dataset_bands: list[HLSBands | int] | None = None,
        output_bands: list[HLSBands | int] | None = None,
        constant_scale: float = 1,
        rgb_indices: list[int] | None = None,
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        expand_temporal_dimension: bool = False,
        reduce_zero_label: bool = False,
        no_data_replace: float | None = None,
        no_label_replace: int | None = None,
        drop_last: bool = True,
        **kwargs: Any,
    ) -> None:
        """Constructor

        Args:
            batch_size (int): _description_
            num_workers (int): _description_
            train_data_root (Path): _description_
            val_data_root (Path): _description_
            test_data_root (Path): _description_
            predict_data_root (Path): _description_
            img_grep (str): _description_
            label_grep (str): _description_
            means (list[float]): _description_
            stds (list[float]): _description_
            train_label_data_root (Path | None, optional): _description_. Defaults to None.
            val_label_data_root (Path | None, optional): _description_. Defaults to None.
            test_label_data_root (Path | None, optional): _description_. Defaults to None.
            train_split (Path | None, optional): _description_. Defaults to None.
            val_split (Path | None, optional): _description_. Defaults to None.
            test_split (Path | None, optional): _description_. Defaults to None.
            ignore_split_file_extensions (bool, optional): Whether to disregard extensions when using the split
                file to determine which files to include in the dataset.
                E.g. necessary for Eurosat, since the split files specify ".jpg" but files are
                actually ".jpg". Defaults to True.
            allow_substring_split_file (bool, optional): Whether the split files contain substrings
                that must be present in file names to be included (as in mmsegmentation), or exact
                matches (e.g. eurosat). Defaults to True.
            dataset_bands (list[HLSBands | int] | None, optional): _description_. Defaults to None.
            predict_dataset_bands (list[HLSBands | int] | None, optional): _description_. Defaults to None.
            output_bands (list[HLSBands | int] | None, optional): _description_. Defaults to None.
            constant_scale (float, optional): _description_. Defaults to 1.
            rgb_indices (list[int] | None, optional): _description_. Defaults to None.
            train_transform (Albumentations.Compose | None): Albumentations transform
                to be applied to the train dataset.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            val_transform (Albumentations.Compose | None): Albumentations transform
                to be applied to the train dataset.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            test_transform (Albumentations.Compose | None): Albumentations transform
                to be applied to the train dataset.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            no_data_replace (float | None): Replace nan values in input images with this value. If none, does no replacement. Defaults to None.
            no_label_replace (int | None): Replace nan values in label with this value. If none, does no replacement. Defaults to None.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Defaults to False.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
            drop_last (bool): Drop the last batch if it is not complete. Defaults to True.
        """
        super().__init__(GenericNonGeoPixelwiseRegressionDataset, batch_size, num_workers, **kwargs)
        self.img_grep = img_grep
        self.label_grep = label_grep
        self.train_root = train_data_root
        self.val_root = val_data_root
        self.test_root = test_data_root
        self.predict_root = predict_data_root
        self.train_split = train_split
        self.val_split = val_split
        self.test_split = test_split
        self.ignore_split_file_extensions = ignore_split_file_extensions
        self.allow_substring_split_file = allow_substring_split_file
        self.drop_last = drop_last
        self.expand_temporal_dimension = expand_temporal_dimension
        self.reduce_zero_label = reduce_zero_label

        self.train_label_data_root = train_label_data_root
        self.val_label_data_root = val_label_data_root
        self.test_label_data_root = test_label_data_root

        self.constant_scale = constant_scale

        self.dataset_bands = dataset_bands
        self.predict_dataset_bands = predict_dataset_bands if predict_dataset_bands else dataset_bands
        self.output_bands = output_bands
        self.rgb_indices = rgb_indices

        # self.aug = AugmentationSequential(
        #     K.Normalize(means, stds),
        #     data_keys=["image"],
        # )
        self.aug = Normalize(means, stds)
        self.no_data_replace = no_data_replace
        self.no_label_replace = no_label_replace

        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)

    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(
                self.train_root,
                image_grep=self.img_grep,
                label_grep=self.label_grep,
                label_data_root=self.train_label_data_root,
                split=self.train_split,
                ignore_split_file_extensions=self.ignore_split_file_extensions,
                allow_substring_split_file=self.allow_substring_split_file,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                rgb_indices=self.rgb_indices,
                transform=self.train_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                self.val_root,
                image_grep=self.img_grep,
                label_grep=self.label_grep,
                label_data_root=self.val_label_data_root,
                split=self.val_split,
                ignore_split_file_extensions=self.ignore_split_file_extensions,
                allow_substring_split_file=self.allow_substring_split_file,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                rgb_indices=self.rgb_indices,
                transform=self.val_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                self.test_root,
                image_grep=self.img_grep,
                label_grep=self.label_grep,
                label_data_root=self.test_label_data_root,
                split=self.test_split,
                ignore_split_file_extensions=self.ignore_split_file_extensions,
                allow_substring_split_file=self.allow_substring_split_file,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                rgb_indices=self.rgb_indices,
                transform=self.test_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
            )

        if stage in ["predict"] and self.predict_root:
            self.predict_dataset = self.dataset_class(
                self.predict_root,
                dataset_bands=self.predict_dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                rgb_indices=self.rgb_indices,
                transform=self.test_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
            )

    def _dataloader_factory(self, split: str) -> DataLoader[dict[str, Tensor]]:
        """Implement one or more PyTorch DataLoaders.

        Args:
            split: Either 'train', 'val', 'test', or 'predict'.

        Returns:
            A collection of data loaders specifying samples.

        Raises:
            MisconfigurationException: If :meth:`setup` does not define a
                dataset or sampler, or if the dataset or sampler has length 0.
        """
        dataset = self._valid_attribute(f"{split}_dataset", "dataset")
        batch_size = self._valid_attribute(f"{split}_batch_size", "batch_size")
        return DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=split == "train",
            num_workers=self.num_workers,
            collate_fn=self.collate_fn,
            drop_last=split == "train" and self.drop_last,
        )



from terratorch.datasets.fire_scars import FireScarsHLS, FireScarsNonGeo, FireScarsSegmentationMask
from terratorch.datasets.generic_pixel_wise_dataset import (
    GenericNonGeoPixelwiseRegressionDataset,
    GenericNonGeoSegmentationDataset,
)
from terratorch.datasets.generic_scalar_label_dataset import (
    GenericNonGeoClassificationDataset,
)

# GenericNonGeoRegressionDataset,
from terratorch.datasets.sen1floods11 import Sen1Floods11NonGeo
from terratorch.datasets.utils import HLSBands

# TorchGeo RasterDatasets
from terratorch.datasets.wsf import WSF2019, WSFEvolution
from terratorch.datasets.hls import HLSL30, HLSS30

# geobench datasets classification
from terratorch.datasets.m_eurosat import MEuroSATNonGeo
from terratorch.datasets.m_bigearthnet import MBigEarthNonGeo
from terratorch.datasets.m_brick_kiln import MBrickKilnNonGeo
from terratorch.datasets.m_forestnet import MForestNetNonGeo
from terratorch.datasets.m_so2sat import MSo2SatNonGeo
from terratorch.datasets.m_pv4ger import MPv4gerNonGeo

# geobench datasets segmentation
from terratorch.datasets.m_cashew_plantation import MBeninSmallHolderCashewsNonGeo
from terratorch.datasets.m_nz_cattle import MNzCattleNonGeo
from terratorch.datasets.m_chesapeake_landcover import MChesapeakeLandcoverNonGeo
from terratorch.datasets.m_pv4ger_seg import MPv4gerSegNonGeo
from terratorch.datasets.m_SA_crop_type import MSACropTypeNonGeo
from terratorch.datasets.m_neontree import MNeonTreeNonGeo

__all__ = (
    "GenericNonGeoSegmentationDataset",
    "GenericNonGeoPixelwiseRegressionDataset",
    "GenericNonGeoClassificationDataset",
    "GenericNonGeoRegressionDataset",
    "FireScarsNonGeo",
    "FireScarsHLS",
    "FireScarsSegmentationMask",
    "Sen1Floods11NonGeo",
    "HLSBands",
    "MEuroSATNonGeo",
    "MBigEarthNonGeo",
    "MBrickKilnNonGeo",
    "MForestNetNonGeo",
    "MSo2SatNonGeo",
    "MPv4gerNonGeo",
    "MBeninSmallHolderCashewsNonGeo",
    "MNzCattleNonGeo",
    "MChesapeakeLandcoverNonGeo",
    "MPv4gerSegNonGeo",
    "MSACropTypeNonGeo",
    "MNeonTreeNonGeo",
    "WSF2019",
    "WSFEvolution",
    "HLSL30",
    "HLSS30",
)
